import datetime
from pychess.Utils.const import RUNNING


class LoadingError(Exception):
    pass


class ChessFile:
    """This class describes an opened chessfile.
    It is lazy in the sense of not parsing any games,
    that the user don't request.
    It has no catching."""

    def __init__(self, file):
        self.file = file
        self.path = file.name if hasattr(file, "name") else None
        self.games = []
        self.offs_ply = {}

    def close(self):
        try:
            self.file.close()
        except OSError:
            pass

    def set_tags_filter(self, text):
        pass

    def set_fen_filter(self, fen):
        pass

    def set_scout_filter(self, query):
        pass

    def get_id(self, gameno):
        return gameno

    def get_records(self, direction=0):
        return self.games, self.offs_ply

    def loadToModel(self, gameno, position, model=None):
        """Load the data of game "gameno" into the gamemodel
        If no model is specified, a new one will be created, loaded and
        returned"""
        raise NotImplementedError

    def __len__(self):
        return len(self.games)

    def get_player_names(self, gameno):
        """Returns a tuple of the players names
        Default is ("Unknown", "Unknown") if nothing is specified"""
        return ("Unknown", "Unknown")

    def get_elo(self, gameno):
        """Returns a tuple of the players rating in ELO format
        Default is 1600 if nothing is specified in the file"""
        return (1600, 1600)

    def get_date(self, gameno):
        """Returns a tuple (year,month,day) of the game date
        Default is current time if nothing is specified in the file"""
        today = datetime.date.today()
        return today.timetuple()[:3]

    def get_site(self, gameno):
        """Returns the location at which the game took place
        Default is "?" if nothing is specified in the file"""
        return "?"

    def get_event(self, gameno):
        """Returns the event at which the game took place
        Could be "World Chess Cup" or "My local tournament"
        Default is "?" if nothing is specified in the file"""
        return "?"

    def get_round(self, gameno):
        """Returns the round of the event at which the game took place
        Pgn supports having subrounds like 2.1.5,
        but as of writing, only the first int is returned.
        Default is 1 if nothing is specified in the file"""
        return 1

    def get_result(self, gameno):
        """Returns the result of the game
        Can be any of: RUNNING, DRAW, WHITEWON or BLACKWON
        Default is RUNNING if nothing is specified in the file"""
        return RUNNING

    def get_variant(self, gameno):
        return 0

    def get_book_moves(self, fen=None):
        return []

    def get_info(self, gameno):
        return None



from gi.repository import GObject

from pychess.ic.icc import DG_ILLEGAL_MOVE
from pychess.ic.managers.ErrorManager import ErrorManager


class ICCErrorManager(ErrorManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(DG_ILLEGAL_MOVE, self.on_icc_illegal_move)

        self.connection.client.run_command("set-2 %s 1" % DG_ILLEGAL_MOVE)

    def on_icc_illegal_move(self, data):
        # gamenumber movestring reason
        gameid, move, reason = data.split(" ", 2)
        self.emit("onIllegalMove", move)



from gi.repository import Gtk

from pychess.System.prefix import addDataPrefix

from .__init__ import CENTER, TabReceiver

from .PyDockComposite import PyDockComposite
from .StarArrowButton import StarArrowButton
from .HighlightArea import HighlightArea


class PyDockLeaf(TabReceiver):
    def __init__(self, widget, title, id, perspective):
        TabReceiver.__init__(self, perspective)
        self.perspective = perspective
        self.set_no_show_all(True)

        self.book = Gtk.Notebook()
        self.book.set_name(id)

        self.book_cids = [
            self.book.connect("drag-begin", self.__onDragBegin),
            self.book.connect("drag-end", self.__onDragEnd),
            self.book.connect_after("switch-page", self.__onPageSwitched),
        ]
        self.add(self.book)
        self.book.show()

        self.highlightArea = HighlightArea(self)

        self.button_cids = []

        self.starButton = StarArrowButton(
            self,
            addDataPrefix("glade/dock_top.svg"),
            addDataPrefix("glade/dock_right.svg"),
            addDataPrefix("glade/dock_bottom.svg"),
            addDataPrefix("glade/dock_left.svg"),
            addDataPrefix("glade/dock_center.svg"),
            addDataPrefix("glade/dock_star.svg"),
        )

        self.button_cids += [
            self.starButton.connect("dropped", self.__onDrop),
            self.starButton.connect("hovered", self.__onHover),
            self.starButton.connect("left", self.__onLeave),
        ]
        self.dockable = True
        self.panels = []

        self.zoomPointer = Gtk.Label()
        self.realtop = None
        self.zoomed = False

        # assert isinstance(widget, Gtk.Notebook)

        self.__add(widget, title, id)

    def _del(self):
        if self.highlightArea.handler_is_connected(self.highlightArea.cid):
            self.highlightArea.disconnect(self.highlightArea.cid)

        for cid in self.button_cids:
            if self.starButton.handler_is_connected(cid):
                self.starButton.disconnect(cid)
        self.button_cids = []

        for cid in self.book_cids:
            if self.book.handler_is_connected(cid):
                self.book.disconnect(cid)

        self.starButton.myparent = None
        self.highlightArea.myparent = None

        TabReceiver._del(self)

    def __repr__(self):
        s = "leaf"  # PyDockLeaf.__repr__(self)
        panels = []
        for widget, title, id in self.getPanels():
            panels.append(id)
        return s + " (" + ", ".join(panels) + ")"

    def __add(self, widget, title, id):
        self.panels.append((widget, title, id))
        self.book.append_page(widget, title)
        self.book.set_tab_detachable(widget, True)
        self.book.set_tab_reorderable(widget, True)
        widget.show_all()

    def dock(self, widget, position, title, id):
        """if position == CENTER: Add a new widget to the leaf-notebook
        if position != CENTER: Fork this leaf into two"""

        if position == CENTER:
            self.__add(widget, title, id)
            return self
        else:
            parent = self.get_parent()
            while not isinstance(parent, PyDockComposite):
                parent = parent.get_parent()

            leaf = PyDockLeaf(widget, title, id, self.perspective)
            new = PyDockComposite(position, self.perspective)
            parent.changeComponent(self, new)
            new.initChildren(self, leaf)
            new.show_all()
            return leaf

    def undock(self, widget):
        """remove the widget from the leaf-notebook
        if this was the only widget, remove this leaf from its owner"""

        gtk_version = (Gtk.get_major_version(), Gtk.get_minor_version())
        if gtk_version >= (3, 16):
            self.book.detach_tab(widget)
        else:
            # To not destroy accidentally our panel widget we need to add a reference to it
            # https://lazka.github.io/pgi-docs/#Gtk-3.0/classes/Container.html#Gtk.Container.remove
            widget._ref()
            self.book.remove(widget)

        for i, (widget_, title, id) in enumerate(self.panels):
            if widget_ == widget:
                break
        else:
            raise KeyError(f"No {widget} in {self}")
        del self.panels[i]

        if self.book.get_n_pages() == 0:
            parent = self.get_parent()
            while not isinstance(parent, PyDockComposite):
                parent = parent.get_parent()
            parent.removeComponent(self)
            self._del()

        return title, id

    def zoomUp(self):
        if self.zoomed:
            return

        from .PyDockTop import PyDockTop

        parent = self.get_parent()
        if not isinstance(parent, PyDockTop):
            while not isinstance(parent, PyDockComposite):
                parent = parent.get_parent()

            parent.changeComponent(self, self.zoomPointer)

            while not isinstance(parent, PyDockTop):
                parent = parent.get_parent()

            self.realtop = parent.getComponents()[0]
            parent.changeComponent(self.realtop, self)

        self.zoomed = True
        self.book.set_show_border(False)

    def zoomDown(self):
        if not self.zoomed:
            return

        if self.zoomPointer.get_parent():
            top_parent = self.get_parent()
            old_parent = self.zoomPointer.get_parent()

            while not isinstance(old_parent, PyDockComposite):
                old_parent = old_parent.get_parent()

            top_parent.changeComponent(self, self.realtop)
            old_parent.changeComponent(self.zoomPointer, self)

        self.realtop = None
        self.zoomed = False
        self.book.set_show_border(True)

    def getPanels(self):
        """Returns a list of (widget, title, id) tuples"""
        return self.panels

    def getCurrentPanel(self):
        for i, (widget, title, id) in enumerate(self.panels):
            if i == self.book.get_current_page():
                return id

    def setCurrentPanel(self, id):
        """Returns the panel id currently shown"""
        for i, (widget, title, id_) in enumerate(self.panels):
            if id == id_:
                self.book.set_current_page(i)
                break

    def isDockable(self):
        return self.dockable

    def setDockable(self, dockable):
        """If the leaf is not dockable it won't be moveable and won't accept
        new panels"""
        self.book.set_show_tabs(dockable)
        # self.book.set_show_border(dockable)
        self.dockable = dockable

    def showArrows(self):
        if self.dockable:
            self.starButton._calcSize()
            self.starButton.show()

    def hideArrows(self):
        self.starButton.hide()
        self.highlightArea.hide()

    def __onDragBegin(self, widget, context):
        for instance in self.getInstances(self.perspective):
            instance.showArrows()

    def __onDragEnd(self, widget, context):
        for instance in self.getInstances(self.perspective):
            instance.hideArrows()

    def __onDrop(self, starButton, position, sender):
        self.highlightArea.hide()

        # if the undocked leaf was alone, __onDragEnd may not triggered
        # because leaf was removed
        for instance in self.getInstances(self.perspective):
            instance.hideArrows()

        if self.dockable:
            if sender.get_parent() == self and self.book.get_n_pages() == 1:
                return
            # cp = sender.get_current_page()
            child = sender.get_nth_page(sender.get_current_page())
            title, id = sender.get_parent().undock(child)
            self.dock(child, position, title, id)

    def __onHover(self, starButton, position, widget):
        if self.dockable:
            self.highlightArea.showAt(position)
            starButton.get_window().raise_()

    def __onLeave(self, starButton):
        self.highlightArea.hide()

    def __onPageSwitched(self, book, page, page_num):
        # When a tab is dragged over another tab, the page is temporally
        # switched, and the notebook child is hovered. Thus we need to reraise
        # our star
        if self.starButton.get_window():
            self.starButton.get_window().raise_()



from gi.repository import GObject

from pychess.ic.FICSObjects import FICSGame
from pychess.ic.managers.HelperManager import HelperManager
from pychess.ic import parseRating, GAME_TYPES_BY_SHORT_FICS_NAME, IC_STATUS_PLAYING
from pychess.ic.icc import (
    DG_PLAYER_ARRIVED_SIMPLE,
    DG_PLAYER_LEFT,
    DG_TOURNEY,
    CN_GAMES,
    DG_WILD_KEY,
)


class ICCHelperManager(HelperManager):
    def __init__(self, helperconn, connection):
        GObject.GObject.__init__(self)

        self.connection = connection

        self.connection.expect_dg_line(
            DG_PLAYER_ARRIVED_SIMPLE, self.on_icc_player_arrived_simple
        )
        self.connection.expect_dg_line(DG_PLAYER_LEFT, self.on_icc_player_left)
        self.connection.expect_dg_line(DG_TOURNEY, self.on_icc_tourney)
        self.connection.expect_dg_line(DG_WILD_KEY, self.on_icc_wild_key)
        self.connection.expect_cn_line(CN_GAMES, self.on_icc_games)

        self.connection.client.run_command("set-2 %s 1" % DG_PLAYER_ARRIVED_SIMPLE)
        self.connection.client.run_command("set-2 %s 1" % DG_PLAYER_LEFT)
        self.connection.client.run_command("set-2 %s 1" % DG_TOURNEY)
        self.connection.client.run_command("set-2 %s 1" % DG_WILD_KEY)

        # From https://www.chessclub.com/user/resources/formats/formats.txt
        # Here is the list of verbose DGs:
        # DG_PLAYER_ARRIVED DG_PLAYER_LEFT
        # DG_GAME_STARTED DG_GAME_RESULT DG_EXAMINED_GAME_IS_GONE
        # DG_PEOPLE_IN_MY_CHANNEL DG_CHANNELS_SHARED DG_SEES_SHOUTS
        # Currently, only TDs like Tomato can use these.

        # Unfortunately we can't maintain full list of ongoing games so we will
        # periodically update top games in ICLounge similar to other ICC clients
        self.connection.client.run_command("games *19")

    def on_icc_games(self, data):
        # 1267      guest2504            1400 KQkr(C)              20u  5  12       W:  1
        # 1060      guest7400            1489 DeadGuyKai            bu  3   0       W: 21
        # 791 1506 PlotinusRedux             guest3090             bu  2  12       W: 21
        # 47 2357 *IM_Danchevski       2683 *GM_Morozevich       Ex: scratch      W: 35
        # 101      Replayer2                 Replayer2            Ex: scratch      W:  1
        # 117 2760 *GM_Topalov          2823 *GM_Caruana          Ex: StLouis16 %0 W: 29
        # 119 1919 stansai              2068 Agrimont             Ex: continuation W: 53
        # 456 games displayed (282 played, 174 examined).
        previous_games = list(self.connection.games.values())
        games = []
        games_got = []
        lines = data.split("\n")
        for line in lines:
            # print(line)
            try:
                parts = line.split()
                index = 0

                gameno = parts[index]
                index += 1

                if parts[index].isdigit():
                    wrating = parts[index]
                    index += 1
                else:
                    wrating = "----"

                wname = parts[index]
                index += 1

                if parts[index].isdigit():
                    brating = parts[index]
                    index += 1
                else:
                    brating = "----"

                bname = parts[index]
                index += 1

                if parts[index] == "Ex:":
                    shorttype = "e"
                    rated = ""
                    min = "0"
                    inc = "0"
                else:
                    rated = parts[index][-1]
                    shorttype = parts[index][:-1]
                    index += 1
                    min = parts[index]
                    index += 1
                    inc = parts[index]
                private = ""
            except IndexError:
                continue

            try:
                gametype = GAME_TYPES_BY_SHORT_FICS_NAME[shorttype]
            except KeyError:
                # TODO: handle ICC wild types
                # print("key error in GAME_TYPES_BY_SHORT_FICS_NAME: %s" % shorttype)
                continue

            wplayer = self.connection.players.get(wname)
            bplayer = self.connection.players.get(bname)
            game = FICSGame(
                wplayer,
                bplayer,
                gameno=int(gameno),
                rated=(rated == "r"),
                private=(private == "p"),
                minutes=int(min),
                inc=int(inc),
                game_type=gametype,
            )

            for player, rating in ((wplayer, wrating), (bplayer, brating)):
                if player.status != IC_STATUS_PLAYING:
                    player.status = IC_STATUS_PLAYING
                if player.game != game:
                    player.game = game
                rating = parseRating(rating)
                if player.ratings[gametype.rating_type] != rating:
                    player.ratings[gametype.rating_type] = rating
                    player.emit("ratings_changed", gametype.rating_type, player)
            if game not in previous_games:
                game = self.connection.games.get(game, emit=False)
                games.append(game)
            games_got.append(game)

        for game in previous_games:
            if (
                game not in games_got
                and game not in self.connection.bm.gamesImObserving
                and game is not self.connection.bm.theGameImPlaying
            ):
                self.connection.games.game_ended(game)
                game.wplayer.game = None
                game.bplayer.game = None

        self.connection.games.emit("FICSGameCreated", games)

    def on_icc_player_arrived_simple(self, data):
        name = data.split()[0]
        player = self.connection.players.get(name)
        player.online = True

    def on_icc_player_left(self, data):
        name = data.split()[0]
        self.connection.players.player_disconnected(name)

    def on_icc_wild_key(self, data):
        key, name = data.split(" ", 1)
        # print(key, name)

    def on_icc_tourney(self, data):
        # index bitfield description join-command watch-command info-command confirm-text
        # 42 0 {Cooly Over 1500 Sunday Top Player Luton Blitz 5 0 rated Rating: 1500..3000 manager bigcol - 7 rounds Tournament Current round:1 Players:8, Latejoin allowed until round: 4} {tell Cooly latejoin & tell 224 Hi, i am in} {} {tell Cooly info} {Do you want to join the Cooly tournament}
        # 59 0 {Tomato U1600 Scheduled Swiss Blitz 2 5 rated Rating: 0..1599 manager Duguesclin - 7 rounds Tournament Current round:1 Players:13, Latejoin allowed until round: 4} {tell Tomato latejoin & tell 46 Hi, i am in} {} {tell Tomato info} {Do you want to join the Tomato tournament}
        # 64 0 {Yenta The STC Sunday Swiss Luton Standard 45 5 rated manager alonzob - 3 rounds Tournament Current round:2 Players:9, Latejoin allowed until round: 2} {tell Yenta latejoin & tell 232 Hi, i am in} {} {tell Yenta info} {Do you want to join the Yenta tournament}
        # 96 6 {} {} {} {} {}
        # 97 6 {[AUDIO] LIVE commentary with GM Ronen Har-Zvi and GM Alex Yermolinsky} {tell webcast listen} {} {} {}
        # 98 6 {[AUDIO] LIVE Espanol con el GM Jordi Magem (ESP)} {tell webcast espanol} {} {} {}
        # 99 6 {LIVE COVERAGE FIDE World Chess Championship 2016 - Game 7} {} {} {finger WorldChamp16} {}
        # 101 6 {LIVE GM Sergey Karjakin(2772) - GM Magnus Carlsen(2853)} {} {observe 1} {} {}
        # 319 6 {Nov 18-- [VIDEO] GM Ronen Har-Zvi analyzes game 6 of the World Chess Championship Match 2016} {} {https://webcast.chessclub.com/icc/i/WC16/Game5/GOTD.html} {https://www20.chessclub.com/article/fide-wc-match-2016-game-6} {}
        # 320 6 {Nov 20 -- [VIDEO] Attack with LarryC! Making Book with the Rook} {https://webcast.chessclub.com/icc/i/LarryC/2011_09_21/Attack_LarryC.html} {} {http://www.chessclub.com/chessfm/index/larryc/index.html} {}

        # TODO
        pass



# Chess960 (Fischer Random Chess)

from random import randrange
from math import floor

from pychess.Utils.const import FISCHERRANDOMCHESS, VARIANTS_SHUFFLE
from pychess.Utils.Board import Board
from pychess.Utils.const import reprFile


class FischerandomBoard(Board):
    variant = FISCHERRANDOMCHESS
    __desc__ = _(
        "http://en.wikipedia.org/wiki/Chess960\n"
        + "FICS wild/fr: http://www.freechess.org/Help/HelpFiles/wild.html"
    )
    name = _("Fischer Random")
    cecp_name = "fischerandom"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def getFrcFen(self, position=519):
        position = max(1, min(960, position))
        pieces = ["", "", "", "", "", "", "", ""]

        # Bishops
        pieces[floor(0.08 * (floor(25 * (position - 1)) % 100) + 1.5)] = "b"
        pieces[floor(0.08 * (floor(25 * floor((position - 1) / 4)) % 100) + 0.5)] = "b"

        # Queen
        z = floor(floor((position - 1) / 4) / 4) / 6
        p = floor(6 * (z - floor(z)) + 0.5)
        for i in range(8):
            if pieces[i] == "":
                if p == 0:
                    pieces[i] = "q"
                    break
                p -= 1

        # KRN
        krn = [
            "nnrkr",
            "nrnkr",
            "nrknr",
            "nrkrn",
            "rnnkr",
            "rnknr",
            "rnkrn",
            "rknnr",
            "rknrn",
            "rkrnn",
        ][floor(z)]
        for i in range(8):
            if pieces[i] == "":
                pieces[i] = krn[:1]
                krn = krn[1:]

        # Castling
        castling = ""
        for i in range(8):
            if pieces[i] == "r":
                castling += reprFile[i]

        # FEN
        pieces = "".join(pieces)
        return "{}/pppppppp/8/8/8/8/PPPPPPPP/{} w {}{} - 0 1".format(
            pieces, pieces.upper(), castling.upper(), castling
        )

    def shuffle_start(self):
        return self.getFrcFen(randrange(1, 960))


if __name__ == "__main__":
    frcBoard = FischerandomBoard(True)
    for i in range(10):
        print(frcBoard.shuffle_start())



from gi.repository import Gtk

from pychess.Utils.const import UNDOABLE_STATES, PRACTICE_GOAL_REACHED
from pychess.Utils.Cord import Cord
from pychess.Utils.LearnModel import learn2str, LESSON, PUZZLE
from pychess.perspectives.learn.PuzzlesPanel import start_puzzle_from
from pychess.perspectives.learn.EndgamesPanel import start_endgame_from
from pychess.perspectives.learn.LessonsPanel import start_lesson_from
from pychess.widgets import preferencesDialog

HINT, MOVE, RETRY, CONTINUE, BACK_TO_MAINLINE, NEXT, SKIP, PREVIOUS = range(8)


css = """
@define-color error_fg_color rgb (235, 235, 235);
@define-color error_bg_color rgb (223, 56, 44);


.error {
    background-image: -gtk-gradient (linear, left top, left bottom,
                                     from (shade (@error_bg_color, 1.04)),
                                     to (shade (@error_bg_color, 0.96)));
    border-style: solid;
    border-width: 1px;

    color: @error_fg_color;

    border-color: shade (@error_bg_color, 0.8);
    border-bottom-color: shade (@error_bg_color, 0.75);

    box-shadow: inset 1px 0 shade (@error_bg_color, 1.08),
                inset -1px 0 shade (@error_bg_color, 1.08),
                inset 0 1px shade (@error_bg_color, 1.1),
                inset 0 -1px shade (@error_bg_color, 1.04);
}
"""


def add_provider(widget):
    screen = widget.get_screen()
    style = widget.get_style_context()
    provider = Gtk.CssProvider()
    provider.load_from_data(css.encode("utf-8"))
    style.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)


class LearnInfoBar(Gtk.InfoBar):
    def __init__(self, gamemodel, boardcontrol, annotation_panel):
        Gtk.InfoBar.__init__(self)
        self.connect("realize", add_provider)

        self.content_area = self.get_content_area()
        self.action_area = self.get_action_area()

        self.gamemodel = gamemodel
        self.boardcontrol = boardcontrol
        self.boardview = boardcontrol.view
        self.annotation_panel = annotation_panel

        self.gamemodel.connect("game_changed", self.on_game_changed)
        self.gamemodel.connect("goal_checked", self.on_goal_checked)
        self.connect("response", self.on_response)

        self.your_turn()

    def clear(self):
        for item in self.content_area:
            self.content_area.remove(item)

        for item in self.action_area:
            self.action_area.remove(item)

    def your_turn(self, shown_board=None):
        self.clear()
        self.set_message_type(Gtk.MessageType.QUESTION)
        self.content_area.add(Gtk.Label(_("Your turn.")))
        self.add_button(_("Hint"), HINT)
        self.add_button(_("Best move"), MOVE)
        if not self.gamemodel.end_game and self.gamemodel.current_index > 0:
            self.add_button(_("Previous"), PREVIOUS)
        self.add_button(_("Skip"), SKIP)
        self.show_all()

    def get_next_puzzle(self):
        self.clear()
        self.set_message_type(Gtk.MessageType.INFO)
        if (
            self.gamemodel.learn_type in (LESSON, PUZZLE)
            and self.gamemodel.current_index + 1 == self.gamemodel.game_count
        ):
            self.content_area.add(
                Gtk.Label(
                    _("Well done! %s completed." % learn2str[self.gamemodel.learn_type])
                )
            )
        else:
            if "FEN" in self.gamemodel.tags:
                self.content_area.add(Gtk.Label(_("Well done!")))
            self.add_button(_("Next"), NEXT)
        self.show_all()
        preferencesDialog.SoundTab.playAction("puzzleSuccess")

        self.gamemodel.solved = True

    def opp_turn(self):
        self.clear()
        self.set_message_type(Gtk.MessageType.INFO)
        self.add_button(_("Continue"), CONTINUE)

        # disable playing
        self.boardcontrol.game_preview = True
        self.show_all()

    def retry(self):
        self.clear()
        self.set_message_type(Gtk.MessageType.ERROR)
        self.content_area.add(Gtk.Label(_("Not the best move!")))
        self.add_button(_("Retry"), RETRY)

        # disable playing
        self.boardcontrol.game_preview = True

        # disable retry button until engine thinking on next move
        if (
            self.gamemodel.practice_game
            and self.gamemodel.status not in UNDOABLE_STATES
        ):
            self.set_response_sensitive(RETRY, False)
        self.show_all()

    def back_to_mainline(self):
        self.clear()
        self.set_message_type(Gtk.MessageType.INFO)
        self.content_area.add(
            Gtk.Label(_("Cool! Now let see how it goes in the main line."))
        )
        self.add_button(_("Back to main line"), BACK_TO_MAINLINE)

        # disable playing
        self.boardcontrol.game_preview = True
        self.show_all()

    def on_response(self, widget, response):
        if response in (HINT, MOVE):
            if self.gamemodel.lesson_game:
                next_move = self.gamemodel.getMoveAtPly(
                    self.boardview.shown, self.boardview.shown_variation_idx
                )
                hints = {self.boardview.shown: ((next_move.as_uci(),),)}
            else:
                hints = self.gamemodel.hints

            if self.boardview.shown in hints:
                if self.boardview.arrows:
                    self.boardview.arrows.clear()
                if self.boardview.circles:
                    self.boardview.circles.clear()

                hint = hints[self.boardview.shown][0][0]
                cord0 = Cord(hint[0], int(hint[1]), "G")
                cord1 = Cord(hint[2], int(hint[3]), "G")
                if response == HINT:
                    self.boardview.circles.add(cord0)
                    self.boardview.redrawCanvas()
                else:
                    self.boardview.arrows.add((cord0, cord1))
                    self.boardview.redrawCanvas()
            else:
                # TODO:
                print(
                    "No hint available yet!", self.gamemodel.ply, self.boardview.shown
                )

        elif response == RETRY:
            self.your_turn()

            if self.gamemodel.practice_game:
                me_played_last_move = (
                    self.gamemodel.boards[-1].color != self.gamemodel.boards[0].color
                )
                moves = (
                    1
                    if self.gamemodel.status in UNDOABLE_STATES and me_played_last_move
                    else 2
                )
                self.gamemodel.undoMoves(moves)

            elif self.gamemodel.lesson_game:
                prev_board = self.gamemodel.getBoardAtPly(
                    self.boardview.shown - 1,
                    variation=self.boardview.shown_variation_idx,
                )

                self.annotation_panel.choices_enabled = False

                self.boardview.setShownBoard(prev_board)
                # We have to fix show_variation_index here, unless
                # after removing the variation it will be invalid!
                for vari in self.gamemodel.variations:
                    if prev_board in vari:
                        break
                self.boardview.shown_variation_idx = self.gamemodel.variations.index(
                    vari
                )

                self.annotation_panel.choices_enabled = True

            self.boardcontrol.game_preview = False

        elif response == CONTINUE:
            self.your_turn()
            self.boardview.showNext()
            self.boardcontrol.game_preview = False

        elif response == BACK_TO_MAINLINE:
            self.opp_turn()
            self.boardview.backToMainLine()
            self.boardcontrol.game_preview = False

        elif response in (NEXT, SKIP):
            if self.gamemodel.puzzle_game:
                if self.gamemodel.from_lesson:
                    start_lesson_from(
                        self.gamemodel.source, self.gamemodel.current_index + 1
                    )
                else:
                    start_puzzle_from(
                        self.gamemodel.source, self.gamemodel.current_index + 1
                    )
            elif self.gamemodel.end_game:
                start_endgame_from(self.gamemodel.source)
            elif self.gamemodel.lesson_game:
                start_lesson_from(
                    self.gamemodel.source, self.gamemodel.current_index + 1
                )
            else:
                print(self.gamemodel.__dir__())

        elif response == PREVIOUS:
            if self.gamemodel.puzzle_game:
                if self.gamemodel.from_lesson:
                    start_lesson_from(
                        self.gamemodel.source, self.gamemodel.current_index - 1
                    )
                else:
                    start_puzzle_from(
                        self.gamemodel.source, self.gamemodel.current_index - 1
                    )
            elif self.gamemodel.end_game:
                start_endgame_from(self.gamemodel.source)
            elif self.gamemodel.lesson_game:
                start_lesson_from(
                    self.gamemodel.source, self.gamemodel.current_index - 1
                )
            else:
                print(self.gamemodel.__dir__())

    def opp_choice_selected(self, board):
        self.your_turn()
        self.boardcontrol.game_preview = False

    def on_game_changed(self, gamemodel, ply):
        if gamemodel.practice_game:
            if len(gamemodel.moves) % 2 == 0:
                # engine moved, we can enable retry
                self.set_response_sensitive(RETRY, True)
                return

    def on_goal_checked(self, gamemodel):
        if gamemodel.status in UNDOABLE_STATES and self.gamemodel.end_game:
            self.get_next_puzzle()

        elif gamemodel.reason == PRACTICE_GOAL_REACHED:
            self.get_next_puzzle()

        elif gamemodel.failed_playing_best:
            self.retry()

        else:
            self.your_turn()



from pychess.Utils.const import (
    EUROSHOGICHESS,
    VARIANTS_OTHER_NONSTANDARD,
    A8,
    B8,
    C8,
    D8,
    E8,
    F8,
    G8,
    H8,
    A7,
    B7,
    C7,
    D7,
    E7,
    F7,
    G7,
    H7,
    A6,
    B6,
    C6,
    D6,
    E6,
    F6,
    G6,
    H6,
    A1,
    B1,
    C1,
    D1,
    E1,
    F1,
    G1,
    H1,
    A2,
    B2,
    C2,
    D2,
    E2,
    F2,
    G2,
    H2,
    A3,
    B3,
    C3,
    D3,
    E3,
    F3,
    G3,
    H3,
)


from pychess.Utils.Board import Board

EUROSHOGISTART = "1nbqkqn1/1r4b1/pppppppp/8/8/PPPPPPPP/1B4R1/1NQKQBN1 w - - 0 1"


class EuroShogiBoard(Board):
    variant = EUROSHOGICHESS
    __desc__ = _("EuroShogi: http://en.wikipedia.org/wiki/EuroShogi")
    name = _("EuroShogi")
    cecp_name = "euroshogi"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    PROMOTION_ZONE = (
        (
            A8,
            B8,
            C8,
            D8,
            E8,
            F8,
            G8,
            H8,
            A7,
            B7,
            C7,
            D7,
            E7,
            F7,
            G7,
            H7,
            A6,
            B6,
            C6,
            D6,
            E6,
            F6,
            G6,
            H6,
        ),
        (
            A1,
            B1,
            C1,
            D1,
            E1,
            F1,
            G1,
            H1,
            A2,
            B2,
            C2,
            D2,
            E2,
            F2,
            G2,
            H2,
            A3,
            B3,
            C3,
            D3,
            E3,
            F3,
            G3,
            H3,
        ),
    )

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=EUROSHOGISTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import math

from gi.repository import GObject, Gtk, Gdk, Pango, PangoCairo
import cairo

# ceil = lambda f: int(math.ceil(f))


def ceil(f):
    return int(math.ceil(f))


line = 10
curve = 60
dotSmall = 14
dotLarge = 24
lineprc = 1 / 7.0

hpadding = 5
vpadding = 3


class SpotGraph(Gtk.EventBox):
    __gsignals__ = {"spotClicked": (GObject.SignalFlags.RUN_FIRST, None, (str,))}

    def __init__(self):
        GObject.GObject.__init__(self)
        self.connect("draw", self.expose)

        # rgb 85,152,215 = Rated
        # rgb 115,210,22 = Unrated
        # rgb 189,47,26 = Computer
        self.typeColors = [
            [[85, 152, 215], [59, 106, 151]],
            [[115, 210, 22], [78, 154, 6]],
            [[85, 152, 215], [189, 47, 26]],
            [[115, 210, 22], [189, 47, 26]],
        ]
        for type in self.typeColors:
            for color in type:
                color[0] = color[0] / 255.0
                color[1] = color[1] / 255.0
                color[2] = color[2] / 255.0

        self.add_events(
            Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
        )
        self.state = 0
        self.connect("button_press_event", self.button_press)
        self.connect("button_release_event", self.button_release)
        self.connect("motion_notify_event", self.motion_notify)
        self.connect("leave_notify_event", self.motion_notify)
        self.connect("size-allocate", self.size_allocate)

        self.cords = []
        self.hovered = None
        self.pressed = False
        self.spots = {}
        self.spotQueue = []  # For spots added prior to widget allocation

        self.xmarks = []
        self.ymarks = []
        self.set_visible_window(False)

    ############################################################################
    # Drawing                                                                  #
    ############################################################################

    def redraw_canvas(self, prect=None):
        if self.get_window():
            if not prect:
                alloc = self.get_allocation()
                prect = (0, 0, alloc.width, alloc.height)
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = prect
            self.get_window().invalidate_rect(rect, True)
            self.get_window().process_updates(True)

    def expose(self, widget, ctx):
        context = widget.get_window().cairo_create()
        self.draw(context)
        return False

    def draw(self, context):
        alloc = self.get_allocation()
        width = alloc.width
        height = alloc.height

        # ------------------------------------------------------ Paint side ruler
        context.move_to(alloc.x + line, alloc.y + line)
        context.rel_line_to(0, height - line * 2 - curve)
        context.rel_curve_to(0, curve, 0, curve, curve, curve)
        context.rel_line_to(width - line * 2 - curve, 0)

        style_ctxt = self.get_style_context()
        dark_prelight = style_ctxt.lookup_color("p_dark_prelight")[1]
        fg_prelight = style_ctxt.lookup_color("p_fg_prelight")[1]
        bg_prelight = style_ctxt.lookup_color("p_bg_prelight")[1]

        context.set_line_width(line)
        context.set_line_cap(cairo.LINE_CAP_ROUND)
        # state = self.state == Gtk.StateType.NORMAL and Gtk.StateType.PRELIGHT or self.state
        context.set_source_rgba(
            dark_prelight.red,
            dark_prelight.green,
            dark_prelight.blue,
            dark_prelight.alpha,
        )
        context.stroke()

        # ------------------------------------------------ Paint horizontal marks
        for x_loc, title in self.xmarks:
            context.set_source_rgba(
                fg_prelight.red, fg_prelight.green, fg_prelight.blue, fg_prelight.alpha
            )
            context.set_font_size(12)
            x_loc, y_loc = self.prcToPix(x_loc, 1)
            context.move_to(x_loc + line / 2.0, y_loc - line / 2.0)
            context.rotate(-math.pi / 2)
            context.show_text(title)
            context.rotate(math.pi / 2)

            context.set_source_rgba(
                bg_prelight.red, bg_prelight.green, bg_prelight.blue, bg_prelight.alpha
            )
            context.move_to(x_loc - line / 2.0, y_loc)
            context.rel_curve_to(6, 0, 6, line, 6, line)
            context.rel_curve_to(0, -line, 6, -line, 6, -line)
            context.close_path()
            context.fill()

        # -------------------------------------------------- Paint vertical marks
        for y_loc, title in self.ymarks:
            context.set_source_rgba(
                fg_prelight.red, fg_prelight.green, fg_prelight.blue, fg_prelight.alpha
            )
            context.set_font_size(12)
            x_loc, y_loc = self.prcToPix(0, y_loc)
            context.move_to(x_loc + line / 2.0, y_loc + line / 2.0)
            context.show_text(title)

            context.set_source_rgba(
                bg_prelight.red, bg_prelight.green, bg_prelight.blue, bg_prelight.alpha
            )
            context.move_to(x_loc, y_loc - line / 2.0)
            context.rel_curve_to(0, 6, -line, 6, -line, 6)
            context.rel_curve_to(line, 0, line, 6, line, 6)
            context.close_path()
            context.fill()

        # ----------------------------------------------------------- Paint spots
        context.set_line_width(dotSmall * lineprc)
        for x_loc, y_loc, col_type, name, text in self.spots.values():
            context.set_source_rgb(*self.typeColors[col_type][0])
            if self.hovered and name == self.hovered[3]:
                continue

            x_loc, y_loc = self.prcToPix(x_loc, y_loc)
            context.arc(x_loc, y_loc, dotSmall / (1 + lineprc) / 2.0, 0, 2 * math.pi)
            context.fill_preserve()
            context.set_source_rgb(*self.typeColors[col_type][1])
            context.stroke()

        # --------------------------------------------------- Paint hovered spots
        context.set_line_width(dotLarge * lineprc)
        if self.hovered:
            x_loc, y_loc, col_type, name, text = self.hovered
            x_loc, y_loc = self.prcToPix(x_loc, y_loc)
            if not self.pressed:
                context.set_source_rgb(*self.typeColors[col_type][0])
            else:
                context.set_source_rgb(*self.typeColors[col_type][1])
            context.arc(x_loc, y_loc, dotLarge / (1 + lineprc) / 2.0, 0, 2 * math.pi)
            context.fill_preserve()
            context.set_source_rgb(*self.typeColors[col_type][1])
            context.stroke()

            x_loc, y_loc, width, height = self.getTextBounds(self.hovered)

            style_ctxt = self.get_style_context()
            style_ctxt.save()
            style_ctxt.add_class(Gtk.STYLE_CLASS_NOTEBOOK)
            Gtk.render_background(
                style_ctxt,
                context,
                int(x_loc - hpadding),
                int(y_loc - vpadding),
                ceil(width + hpadding * 2),
                ceil(height + vpadding * 2),
            )
            Gtk.render_frame(
                style_ctxt,
                context,
                int(x_loc - hpadding),
                int(y_loc - vpadding),
                ceil(width + hpadding * 2),
                ceil(height + vpadding * 2),
            )
            style_ctxt.restore()

            context.move_to(x_loc, y_loc)
            context.set_source_rgba(
                fg_prelight.red, fg_prelight.green, fg_prelight.blue, fg_prelight.alpha
            )
            PangoCairo.show_layout(context, self.create_pango_layout(text))

    ############################################################################
    # Events                                                                   #
    ############################################################################

    def button_press(self, widget, event):
        alloc = self.get_allocation()
        self.cords = [event.x + alloc.x, event.y + alloc.y]
        self.pressed = True
        if self.hovered:
            self.redraw_canvas(self.getBounds(self.hovered))

    def button_release(self, widget, event):
        alloc = self.get_allocation()
        self.cords = [event.x + alloc.x, event.y + alloc.y]
        self.pressed = False
        if self.hovered:
            self.redraw_canvas(self.getBounds(self.hovered))
            if event.button == 1 and self.pointIsOnSpot(
                event.x + alloc.x, event.y + alloc.y, self.hovered
            ):
                self.emit("spotClicked", self.hovered[3])

    def motion_notify(self, widget, event):
        alloc = self.get_allocation()
        self.cords = [event.x + alloc.x, event.y + alloc.y]
        spot = self.getSpotAtPoint(*self.cords)
        if self.hovered and spot == self.hovered:
            return
        if self.hovered:
            bounds = self.getBounds(self.hovered)
            self.hovered = None
            self.redraw_canvas(bounds)
        if spot:
            self.hovered = spot
            self.redraw_canvas(self.getBounds(self.hovered))

    def size_allocate(self, widget, allocation):
        assert self.get_allocation().width > 1
        for spot in self.spotQueue:
            self.addSpot(*spot)
        del self.spotQueue[:]

    ############################################################################
    # Interaction                                                              #
    ############################################################################

    def addSpot(self, name, text, x0, y0, colour_type=0):
        """x and y are in % from 0 to 1"""
        assert colour_type in range(len(self.typeColors))

        if self.get_allocation().width <= 1:
            self.spotQueue.append((name, text, x0, y0, colour_type))
            return

        x_loc1, y_loc1 = self.getNearestFreeNeighbourHexigon(x0, 1 - y0)
        spot = (x_loc1, y_loc1, colour_type, name, text)
        self.spots[name] = spot
        if (
            not self.hovered
            and self.cords
            and self.pointIsOnSpot(self.cords[0], self.cords[1], spot)
        ):
            self.hovered = spot
        self.redraw_canvas(self.getBounds(spot))

    def removeSpot(self, name):
        if name not in self.spots:
            return
        spot = self.spots.pop(name)
        bounds = self.getBounds(spot)
        if spot == self.hovered:
            self.hovered = None
        self.redraw_canvas(bounds)

    def clearSpots(self):
        self.hovered = None
        self.spots.clear()
        self.redraw_canvas()

    def addXMark(self, x, title):
        self.xmarks.append((x, title))

    def addYMark(self, y, title):
        self.ymarks.append((1 - y, title))

    ############################################################################
    # Internal stuff                                                           #
    ############################################################################

    def getTextBounds(self, spot):
        x_loc, y_loc, type, name, text = spot
        x_loc, y_loc = self.prcToPix(x_loc, y_loc)

        alloc = self.get_allocation()
        width = alloc.width
        #        height = alloc.height

        extends = self.create_pango_layout(text).get_extents()
        scale = float(Pango.SCALE)
        x_bearing, y_bearing, twidth, theight = [
            extends[1].x / scale,
            extends[1].y / scale,
            extends[1].width / scale,
            extends[1].height / scale,
        ]
        tx_loc = x_loc - x_bearing + dotLarge / 2.0
        ty_loc = y_loc - y_bearing - theight - dotLarge / 2.0

        if (
            tx_loc + twidth > width
            and x_loc - x_bearing - twidth - dotLarge / 2.0 > alloc.x
        ):
            tx_loc = x_loc - x_bearing - twidth - dotLarge / 2.0
        if ty_loc < alloc.y:
            ty_loc = y_loc - y_bearing + dotLarge / 2.0

        return (tx_loc, ty_loc, twidth, theight)

    def join(self, r0, r1):
        x_loc1 = min(r0[0], r1[0])
        x_loc2 = max(r0[0] + r0[2], r1[0] + r1[2])
        y_loc1 = min(r0[1], r1[1])
        y_loc2 = max(r0[1] + r0[3], r1[1] + r1[3])
        return (x_loc1, y_loc1, x_loc2 - x_loc1, y_loc2 - y_loc1)

    def getBounds(self, spot):
        (
            x_loc,
            y_loc,
        ) = (
            spot[0],
            spot[1],
        )
        x_loc, y_loc = self.prcToPix(x_loc, y_loc)

        if spot == self.hovered:
            size = dotLarge
        else:
            size = dotSmall

        bounds = (x_loc - size / 2.0 - 1, y_loc - size / 2.0 - 1, size + 2, size + 2)

        if spot == self.hovered:
            x_loc, y_loc, width, height = self.getTextBounds(spot)
            tbounds = (
                x_loc - hpadding,
                y_loc - vpadding,
                width + hpadding * 2 + 1,
                height + vpadding * 2 + 1,
            )
            return self.join(bounds, tbounds)

        return bounds

    def getNearestFreeNeighbourHexigon(self, xorg, yorg):
        """This method performs an hexigon search for an empty place to put a
        new dot."""

        x_loc, y_loc = self.prcToPix(xorg, yorg)
        # Start by testing current spot
        if self.isEmpty(x_loc, y_loc):
            return xorg, yorg

        directions = [
            (math.cos((i + 2) * math.pi / 3), math.sin((i + 2) * math.pi / 3))
            for i in range(6)
        ]

        level = 1
        while True:
            x_loc += dotSmall
            for delta_x, delta_y in directions:
                for i in range(level):
                    if self.isEmpty(x_loc, y_loc):
                        return self.pixToPrc(x_loc, y_loc)
                    x_loc += delta_x * dotSmall
                    y_loc += delta_y * dotSmall
            level += 1

    def getNearestFreeNeighbourArchi(self, xorg, yorg):
        """This method performs an archimedes-spircal search for an empty
        place to put a new dot.
        http://en.wikipedia.org/wiki/Archimedean_spiral"""

        xorg, yorg = self.prcToPix(xorg, yorg)
        # Start by testing current spot
        if self.isEmpty(xorg, yorg):
            return self.pixToPrc(xorg, yorg)

        radius = 0
        while True:
            # This is an approx to the equation
            # cos((radius-s)/(2pi)) = (radius^2+s^2-1)/(2*radius*s)
            # which gives the next point on the spiral 1 away.
            radius = (
                4 * math.pi**3 * radius
                + radius**2
                + math.sqrt(16 * math.pi**6 + 8 * math.pi**3 * radius + radius**4)
            ) / (4 * math.pi**3 + 2 * radius)

            x_loc = radius * math.cos(radius) / (4 * math.pi) * dotSmall + xorg
            y_loc = radius * math.sin(radius) / (4 * math.pi) * dotSmall + yorg
            if self.isEmpty(x_loc, y_loc):
                return self.pixToPrc(x_loc, y_loc)

    def getNearestFreeNeighbourSquare(self, xorg, yorg):
        """This method performs a spircal search for an empty square to put a
        new dot."""

        up = 2
        right = 1
        down = 1
        left = 2

        x_loc, y_loc = self.prcToPix(xorg, yorg)

        # Start by testing current spot
        if self.isEmpty(x_loc, y_loc):
            return self.pixToPrc(x_loc, y_loc)

        while True:
            for i in range(right):
                x_loc += dotSmall
                if self.isEmpty(x_loc, y_loc):
                    return self.pixToPrc(x_loc, y_loc)

            for i in range(down):
                y_loc += dotSmall
                if self.isEmpty(x_loc, y_loc):
                    return self.pixToPrc(x_loc, y_loc)

            for i in range(left):
                x_loc -= dotSmall
                if self.isEmpty(x_loc, y_loc):
                    return self.pixToPrc(x_loc, y_loc)

            for i in range(up):
                y_loc -= dotSmall
                if self.isEmpty(x_loc, y_loc):
                    return self.pixToPrc(x_loc, y_loc)

            # Grow spiral bounds
            right += 2
            down += 2
            left += 2
            up += 2

    def isEmpty(self, x0, y0):
        """Returns true if a spot placed on (x, y) is inside the graph and not
        intersecting with other spots.
        x and y should be in pixels, not percent"""

        # Make sure spiral search don't put dots outside the graph
        x_loc, y_loc = self.prcToPix(0, 0)
        width, height = self.prcToPix(1, 1)

        if not x_loc <= x0 <= width or not y_loc <= y0 <= height:
            return False

        # Tests if the spot intersects any other spots
        for x_loc1, y_loc1, type, name, text in self.spots.values():
            x_loc1, y_loc1 = self.prcToPix(x_loc1, y_loc1)
            if (x_loc1 - x0) ** 2 + (y_loc1 - y0) ** 2 < dotSmall**2 - 0.1:
                return False

        return True

    def pointIsOnSpot(self, x0, y0, spot):
        """Returns true if (x, y) is inside the spot 'spot'. The size of the
        spot is determined based on its hoverness.
        x and y should be in pixels, not percent"""

        if spot == self.hovered:
            size = dotLarge
        else:
            size = dotSmall

        x_loc1, y_loc1, type, name, text = spot
        x_loc1, y_loc1 = self.prcToPix(x_loc1, y_loc1)
        if (x_loc1 - x0) ** 2 + (y_loc1 - y0) ** 2 <= (size / 2.0) ** 2:
            return True
        return False

    def getSpotAtPoint(self, x, y):
        """Returns the spot embrace (x, y) if any. Otherwise it returns None.
        x and y should be in pixels, not percent"""

        if self.hovered and self.pointIsOnSpot(x, y, self.hovered):
            return self.hovered

        for spot in self.spots.values():
            if spot == self.hovered:
                continue
            if self.pointIsOnSpot(x, y, spot):
                return spot

        return None

    def prcToPix(self, x, y):
        """Translates from 0-1 cords to real world cords"""
        alloc = self.get_allocation()
        return (
            x * (alloc.width - line * 1.5 - dotLarge * 0.5) + line * 1.5 + alloc.x,
            y * (alloc.height - line * 1.5 - dotLarge * 0.5) + dotLarge * 0.5 + alloc.y,
        )

    def pixToPrc(self, x, y):
        """Translates from real world cords to 0-1 cords"""
        alloc = self.get_allocation()
        return (x - line * 1.5 - alloc.x) / (
            alloc.width - line * 1.5 - dotLarge * 0.5
        ), (y - dotLarge * 0.5 - alloc.y) / (alloc.height - line * 1.5 - dotLarge * 0.5)


if __name__ == "__main__":
    window = Gtk.Window()

    style_ctx = window.get_style_context()
    data = "@define-color p_bg_color #ededed; \
            @define-color p_light_color #ffffff; \
            @define-color p_dark_color #a6a6a6; \
            @define-color p_dark_prelight #a9a9a9; \
            @define-color p_fg_prelight #313739; \
            @define-color p_bg_prelight #ededed; \
            @define-color p_bg_active #d6d6d6;"

    provider = Gtk.CssProvider.new()
    provider.load_from_data(data)
    style_ctx.add_provider_for_screen(
        Gdk.Screen.get_default(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )

    note_book = Gtk.Notebook()
    window.add(note_book)
    v_box = Gtk.VBox()
    note_book.append_page(v_box, None)

    spot_graph = SpotGraph()
    spot_graph.addXMark(0.5, "Center")
    spot_graph.addYMark(0.5, "Center")
    v_box.pack_start(spot_graph, True, True, 0)

    button = Gtk.Button("New Spot")

    def callback(button):
        if not hasattr(button, "nextnum"):
            button.nextnum = 0
        else:
            button.nextnum += 1
        spot_graph.addSpot(str(button.nextnum), "Blablabla", 1, 1, 0)

    button.connect("clicked", callback)
    v_box.pack_start(button, False, True, 0)

    window.connect("delete-event", Gtk.main_quit)
    window.show_all()
    window.resize(400, 400)
    Gtk.main()



import io
import os
import sys
import subprocess
from urllib.request import url2pathname

from pychess.System.Log import log


class Player:
    def __init__(self):
        self.ready = False

    def play(self, uri):
        pass


sound_player = Player()

if os.environ.get("PYCHESS_UNITTEST"):  # prevent sounds in unittests
    pass

elif sys.platform == "win32":
    import winsound

    class WinsoundPlayer(Player):
        def __init__(self):
            self.ready = True

        def play(self, uri):
            try:
                winsound.PlaySound(None, 0)
                winsound.PlaySound(
                    url2pathname(uri[5:]), winsound.SND_FILENAME | winsound.SND_ASYNC
                )
            except RuntimeError:
                log.error(
                    "ERROR: RuntimeError while playing %s." % url2pathname(uri[5:])
                )

    sound_player = WinsoundPlayer()
else:

    class GstPlayer(Player):
        def __init__(self):
            PYTHONBIN = sys.executable.split("/")[-1]
            try:
                if getattr(sys, "frozen", False):
                    gst_player = os.path.join(
                        os.path.abspath(os.path.dirname(sys.executable)),
                        "gst_player.py",
                    )
                else:
                    gst_player = os.path.join(
                        os.path.abspath(os.path.dirname(__file__)), "gst_player.py"
                    )
                self.player = subprocess.Popen(
                    [PYTHONBIN, gst_player], stdin=subprocess.PIPE
                )

                self.stdin = io.TextIOWrapper(
                    self.player.stdin, encoding="utf-8", line_buffering=True
                )
                self.ready = True
            except Exception:
                self.player = None
                log.error("ERROR: starting gst_player failed")
                raise

        def play(self, uri):
            if self.player is not None:
                try:
                    self.stdin.write(url2pathname(uri[5:]) + "\n")
                except BrokenPipeError:
                    pass

    sound_player = GstPlayer()



import asyncio

from gi.repository import GObject

from pychess.Utils.const import LOCAL, RUNNING
from pychess.Utils.DecisionSupportAlgorithm import DecisionSupportAlgorithm
from pychess.Variants.setupposition import SetupBoard


class SetupMove:
    def __init__(self, move):
        self.cord0 = move[0]
        self.cord1 = move[1]
        self.flag = 0

    def is_capture(self, board):
        return False


class SetupPlayer:
    __type__ = LOCAL

    def __init__(self, board_control):
        self.queue = asyncio.Queue()
        self.board_control = board_control
        self.board_control.connect("action", self.on_action)
        self.board_control.connect("piece_moved", self.piece_moved)

    def on_action(self, bc, action, player, param):
        self.queue.put_nowait((action, param))
        if action == "SETUP":
            # force both virtual player to make_move()
            self.queue.put_nowait((action, param))

    async def make_move(self):
        item = await self.queue.get()
        return item

    def piece_moved(self, board, move, color):
        self.queue.put_nowait((SetupMove(move), color))


class SetupModel(GObject.GObject):
    __gsignals__ = {
        "game_started": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "game_changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "moves_undoing": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "variation_undoing": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "game_loading": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "game_loaded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "game_ended": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.stop = False
        self.lowply = 0
        self.status = RUNNING
        self.players = []
        self.moves = []
        self.variant = SetupBoard
        self.boards = [self.variant()]
        self.variations = [self.boards]

        self.lesson_game = False

        self.support_algorithm = DecisionSupportAlgorithm()

    def _get_ply(self):
        return self.boards[-1].ply

    ply = property(_get_ply)

    def getBoardAtPly(self, ply, variation=0):
        return self.boards[ply]

    def getMoveAtPly(self, ply, variation=0):
        return self.moves[ply]

    def isPlayingICSGame(self):
        # prevent hovering over fields
        return True

    def receiveSupportAlgorithm(self, x):
        return

    def start(self):
        async def coro():
            self.emit("game_started")
            while True:
                player0, player1 = await self.curplayer.make_move()

                if isinstance(player0, SetupMove):
                    # print(player0.cord0, player0.cord1, player1)
                    new_board = self.boards[-1].move(player0, player1)
                    self.moves.append(player0)
                    self.boards.append(new_board)
                    self.emit("game_changed", self.ply)
                elif player0 == "SETUP":
                    # print("SETUP", player0, player1)
                    self.emit("game_ended", 0)
                    self.boards = [self.variant(setup=player1)]
                    self.variations = [self.boards]
                    self.emit("game_loaded", 0)
                    self.emit("game_started")
                    self.emit("game_changed", 0)
                elif player0 == "CLOSE":
                    # print("CLOSE")
                    break

        asyncio.create_task(coro())



from gi.repository import GObject

from pychess.ic.VerboseTelnet import ConsoleHandler


class ConsoleManager(GObject.GObject):
    __gsignals__ = {
        "consoleMessage": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (
                object,
                object,
            ),
        ),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.connection.client.lines.consolehandler = ConsoleHandler(
            self.onConsoleMessage
        )

    def onConsoleMessage(self, lines, ini_lines=None):
        self.emit("consoleMessage", lines, ini_lines)



import json
import os
import re

import pexpect
from pexpect.popen_spawn import PopenSpawn

from pychess.Players.ProtocolEngine import TIME_OUT_SECOND

PGN_HEADERS_REGEX = re.compile(r"\[([A-Za-z0-9_]+)\s+\"(.*)\"\]")


class Scoutfish:
    def __init__(self, engine=""):
        if not engine:
            engine = "./scoutfish"
        self.p = PopenSpawn(engine, timeout=TIME_OUT_SECOND, encoding="utf-8")
        self.wait_ready()
        self.pgn = ""
        self.db = ""

    def wait_ready(self):
        self.p.sendline("isready")
        self.p.expect("readyok")

    def open(self, pgn):
        """Open a PGN file and create an index if not exsisting"""
        if not os.path.isfile(pgn):
            raise NameError(f"File {pgn} does not exsist")
        pgn = os.path.normcase(pgn)
        self.pgn = pgn
        self.db = os.path.splitext(pgn)[0] + ".scout"
        if not os.path.isfile(self.db):
            result = self.make()
            self.db = result["DB file"]

    def close(self):
        """Terminate scoutfish. Not really needed: engine will terminate as
        soon as pipe is closed, i.e. when we exit."""
        self.p.sendline("quit")
        self.p.expect(pexpect.EOF)
        self.pgn = ""
        self.db = ""

    def make(self):
        """Make an index out of a pgn file. Normally called by open()"""
        if not self.pgn:
            raise NameError("Unknown DB, first open a PGN file")
        cmd = "make " + self.pgn
        self.p.sendline(cmd)
        self.wait_ready()
        s = "{" + self.p.before.split("{")[1]
        s = s.replace("\\", r"\\")  # Escape Windows's path delimiter
        result = json.loads(s)
        self.p.before = ""
        return result

    def setoption(self, name, value):
        """Set an option value, like threads number"""
        cmd = f"setoption name {name} value {value}"
        self.p.sendline(cmd)
        self.wait_ready()

    def scout(self, q):
        """Run query defined by 'q' dict. Result will be a dict too"""
        if not self.db:
            raise NameError("Unknown DB, first open a PGN file")
        j = json.dumps(q)
        cmd = f"scout {self.db} {j}"
        self.p.sendline(cmd)
        self.wait_ready()
        result = json.loads(self.p.before)
        self.p.before = ""
        return result

    def scout_raw(self, q):
        """Run query defined by 'q' dict. Result will be full output"""
        if not self.db:
            raise NameError("Unknown DB, first open a PGN file")
        j = json.dumps(q)
        cmd = f"scout {self.db} {j}"
        self.p.sendline(cmd)
        self.wait_ready()
        result = self.p.before
        self.p.before = ""
        return result

    def get_games(self, matches):
        """Retrieve the PGN games specified in the offset list. Games are
        added to each list item with a 'pgn' key"""
        if not self.pgn:
            raise NameError("Unknown DB, first open a PGN file")
        with open(self.pgn) as f:
            for match in matches:
                f.seek(match["ofs"])
                game = ""
                for line in f:
                    if line.startswith('[Event "'):
                        if game:
                            break  # Second one, start of next game
                        else:
                            game = line  # First occurence
                    elif game:
                        game += line
                match["pgn"] = game.strip()
        return matches

    def get_header(self, pgn):
        """Return a dict with just header information out of a pgn game. The
        pgn tags are supposed to be consecutive"""
        header = {}
        for line in pgn.splitlines():
            line = line.strip()
            if line.startswith("[") and line.endswith("]"):
                tag_match = PGN_HEADERS_REGEX.match(line)
                if tag_match:
                    header[tag_match.group(1)] = tag_match.group(2)
            else:
                break
        return header

    def get_game_headers(self, matches):
        """Return a list of headers out of a list of pgn games. It is defined
        to be compatible with the return value of get_games()"""
        headers = []
        for match in matches:
            pgn = match["pgn"]
            h = self.get_header(pgn)
            headers.append(h)
        return headers



import asyncio

from gi.repository import Gtk

from pychess.System import conf
from pychess.Utils.const import COLUMN_ROW_RESET, GTK_ICON_VIEW_REFRESH
from pychess.widgets import mainwindow


def generateLessonsSidepanel(solving_progress, learn_category_id, entries, start_from):
    """
    generateLessonsSidepanel returns a class to be used as a panel by Gtk. More
    specifically, that class is meant to be named Sidepanel.

    generateLessonsSidepanel allows to avoid duplicate code between
    PuzzlesPanel and LessonsPanel.py.
    """

    class Sidepanel:
        def load(self, persp):
            self.persp = persp
            self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

            self.tv = Gtk.TreeView()

            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(_("Title"), renderer, text=1)
            self.tv.append_column(column)

            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(_("Source"), renderer, text=2)
            self.tv.append_column(column)

            renderer = Gtk.CellRendererProgress()
            column = Gtk.TreeViewColumn(_("Progress"), renderer, text=3, value=4)
            column.set_expand(True)
            self.tv.append_column(column)

            renderer = Gtk.CellRendererPixbuf()
            column = Gtk.TreeViewColumn(_("Reset"), renderer, icon_name=5)
            column.set_name(COLUMN_ROW_RESET)
            self.tv.append_column(column)

            self.tv.connect("row-activated", self.row_activated)

            def on_progress_updated(solving_progress, key, progress):
                for i, row in enumerate(self.store):
                    if row[0] == key:
                        (
                            progress_ratio_string,
                            percent,
                            reset_icon,
                        ) = self._compute_progress_info(progress)
                        treeiter = self.store.get_iter(Gtk.TreePath(i))
                        self.store[treeiter][3] = progress_ratio_string
                        self.store[treeiter][4] = percent
                        self.store[treeiter][5] = reset_icon

            solving_progress.connect("progress_updated", on_progress_updated)

            self.store = Gtk.ListStore(str, str, str, str, int, str)

            async def coro():
                for file_name, title, author in entries:
                    progress = solving_progress.get(file_name)
                    (
                        progress_ratio_string,
                        percent,
                        reset_icon,
                    ) = self._compute_progress_info(progress)
                    self.store.append(
                        [
                            file_name,
                            title,
                            author,
                            progress_ratio_string,
                            percent,
                            reset_icon,
                        ]
                    )
                    await asyncio.sleep(0)

            asyncio.create_task(coro())

            self.tv.set_model(self.store)
            self.tv.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
            self.tv.set_cursor(conf.get("learncombo%s" % learn_category_id))

            scrollwin = Gtk.ScrolledWindow()
            scrollwin.add(self.tv)
            scrollwin.show_all()

            self.box.pack_start(scrollwin, True, True, 0)
            self.box.show_all()

            return self.box

        def row_activated(self, widget, path, col):
            if path is None:
                return
            else:
                filename, title, *_ = entries[path[0]]
                if col.get_name() == COLUMN_ROW_RESET:
                    self._reset_progress_file(filename, title)
                else:
                    conf.set("categorycombo", learn_category_id)
                    from pychess.widgets.TaskerManager import learn_tasker

                    learn_tasker.learn_combo.set_active(path[0])
                    start_from(filename)

        @staticmethod
        def _compute_progress_info(progress):
            solved = progress.count(1)
            percent = 0 if solved == 0 else round((solved * 100.0) / len(progress))
            reset_icon = None if solved == 0 else GTK_ICON_VIEW_REFRESH
            return f"{solved} / {len(progress)}", percent, reset_icon

        def _reset_progress_file(self, filename, title):
            progress = solving_progress[filename]
            _str, _percent, reset_icon = self._compute_progress_info(progress)
            if reset_icon is not None:
                dialog = Gtk.MessageDialog(
                    mainwindow(),
                    0,
                    Gtk.MessageType.QUESTION,
                    Gtk.ButtonsType.OK_CANCEL,
                    _(
                        'This will reset the progress to 0 for the puzzle "{title}"'
                    ).format(title=title),
                )
                response = dialog.run()
                if response == Gtk.ResponseType.OK:
                    solving_progress[filename] = [0] * len(progress)
                    self.persp.update_progress(None, None, None)
                dialog.destroy()

    return Sidepanel



import asyncio
from io import StringIO

from gi.repository import Gtk

from pychess.ic.FICSObjects import (
    FICSGame,
    FICSAdjournedGame,
    make_sensitive_if_available,
)
from pychess.ic import get_infobarmessage_content
from pychess.ic.ICGameModel import ICGameModel

from pychess.Utils.const import WHITE, BLACK, REMOTE, reprResult
from pychess.Utils.TimeModel import TimeModel
from pychess.Players.ICPlayer import ICPlayer
from pychess.Savers import pgn
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.perspectives import perspective_manager
from pychess.perspectives.fics.ParrentListSection import (
    ParrentListSection,
    cmp,
    SEPARATOR,
    FOLLOW,
    CHAT,
    CHALLENGE,
    FINGER,
    ARCHIVED,
)
from pychess.widgets.InfoBar import InfoBarMessage, InfoBarMessageButton


__title__ = _("Archived")

__icon__ = addDataPrefix("glade/panel_games.svg")

__desc__ = _("Adjourned, history and journal games list")


class Sidepanel(ParrentListSection):
    def load(self, widgets, connection, lounge):
        self.connection = connection
        self.widgets = widgets
        self.lounge = lounge
        self.infobar = lounge.infobar

        __widget__ = lounge.archive_list

        self.games = {}
        self.messages = {}
        self.tv = widgets["adjournedtreeview"]
        self.store = Gtk.ListStore(
            FICSGame, str, str, str, str, str, str, str, str, str, int
        )
        self.model = Gtk.TreeModelSort(model=self.store)
        self.tv.set_model(self.model)
        self.addColumns(
            self.tv,
            "FICSGame",
            _("White"),
            "",
            "",
            _("Black"),
            "",
            _("Rated"),
            _("Clock"),
            _("Type"),
            _("Date/Time"),
            "sortable_time",
            hide=[0, 10],
        )
        self.selection = self.tv.get_selection()
        self.selection.connect("changed", self.onSelectionChanged)
        self.onSelectionChanged(self.selection)
        self.tv.get_model().set_sort_func(5, self.compareFunction, 7)

        self.connection.adm.connect("adjournedGameAdded", self.onAdjournedGameAdded)
        self.connection.games.connect(
            "FICSAdjournedGameRemoved", self.onAdjournedGameRemoved
        )
        self.connection.adm.connect("historyGameAdded", self.onHistoryGameAdded)
        self.connection.games.connect(
            "FICSHistoryGameRemoved", self.onHistoryGameRemoved
        )
        self.connection.adm.connect("journalGameAdded", self.onJournalGameAdded)
        self.connection.games.connect(
            "FICSJournalGameRemoved", self.onJournalGameRemoved
        )

        widgets["resignButton"].connect("clicked", self.onResignButtonClicked)
        widgets["abortButton"].connect("clicked", self.onAbortButtonClicked)
        widgets["drawButton"].connect("clicked", self.onDrawButtonClicked)
        widgets["resumeButton"].connect("clicked", self.onResumeButtonClicked)
        widgets["previewButton"].connect("clicked", self.onPreviewButtonClicked)
        widgets["examineButton"].connect("clicked", self.onExamineButtonClicked)
        widgets["mygamesButton"].connect("clicked", self.onMygamesButtonClicked)
        self.tv.connect(
            "row-activated", lambda *args: self.onPreviewButtonClicked(None)
        )
        self.connection.bm.connect("archiveGamePreview", self.onGamePreview)
        self.connection.bm.connect("playGameCreated", self.onPlayGameCreated)

        self.tv.connect("button-press-event", self.button_press_event)
        self.createLocalMenu((CHALLENGE, CHAT, FOLLOW, SEPARATOR, FINGER, ARCHIVED))

        return __widget__

    def getSelectedPlayer(self):
        model = self.tv.get_model()
        path, col = self.tv.get_cursor()
        col_index = self.tv.get_columns().index(col)
        game = model.get_value(model.get_iter(path), 0)
        return game.bplayer if col_index >= 3 else game.wplayer

    def onSelectionChanged(self, selection):
        model, treeiter = selection.get_selected()
        a_row_is_selected = False
        if treeiter is not None:
            a_row_is_selected = True
            game = model.get_value(treeiter, 0)
            if (
                isinstance(game, FICSAdjournedGame)
                and self.connection.stored_owner == self.connection.username
            ):
                make_sensitive_if_available(self.widgets["resumeButton"], game.opponent)
                for button in ("resignButton", "abortButton", "drawButton"):
                    self.widgets[button].set_sensitive(True)
            else:
                for button in (
                    "resignButton",
                    "abortButton",
                    "drawButton",
                    "resumeButton",
                ):
                    self.widgets[button].set_sensitive(False)
        else:
            self.widgets["resumeButton"].set_sensitive(False)
            self.widgets["resumeButton"].set_tooltip_text("")
            for button in ("resignButton", "abortButton", "drawButton"):
                self.widgets[button].set_sensitive(False)
        self.widgets["previewButton"].set_sensitive(a_row_is_selected)
        self.widgets["examineButton"].set_sensitive(a_row_is_selected)

    def onPlayGameCreated(self, bm, board):
        for message in self.messages.values():
            message.dismiss()
        self.messages = {}
        return False

    def _infobar_adjourned_message(self, game, player):
        if player not in self.messages:
            text = _(
                " with whom you have an adjourned <b>%(timecontrol)s</b> "
                + "<b>%(gametype)s</b> game is online."
            ) % {
                "timecontrol": game.display_timecontrol,
                "gametype": game.game_type.display_text,
            }
            content = get_infobarmessage_content(player, text, gametype=game.game_type)

            def callback(infobar, response, message):
                log.debug(
                    "%s" % player,
                    extra={
                        "task": (
                            self.connection.username,
                            "_infobar_adjourned_message.callback",
                        )
                    },
                )
                if response == Gtk.ResponseType.ACCEPT:
                    self.connection.client.run_command("match %s" % player.name)
                elif response == Gtk.ResponseType.HELP:
                    self.connection.adm.queryMoves(game)
                else:
                    try:
                        self.messages[player].dismiss()
                        del self.messages[player]
                    except KeyError:
                        pass
                return False

            message = InfoBarMessage(Gtk.MessageType.QUESTION, content, callback)
            message.add_button(
                InfoBarMessageButton(_("Request Continuation"), Gtk.ResponseType.ACCEPT)
            )
            message.add_button(
                InfoBarMessageButton(_("Examine Adjourned Game"), Gtk.ResponseType.HELP)
            )
            message.add_button(
                InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
            )
            make_sensitive_if_available(message.buttons[0], player)
            self.messages[player] = message
            self.infobar.push_message(message)

    def compareFunction(self, treemodel, iter0, iter1, column):
        (minute0, minute1) = (
            treemodel.get_value(iter0, 10),
            treemodel.get_value(iter1, 10),
        )
        return cmp(minute0, minute1)

    def online_changed(self, player, prop, game):
        log.debug(f"AdjournedTabSection.online_changed: {repr(player)} {repr(game)}")
        partner = game.bplayer if game.wplayer.name == player.name else game.wplayer
        result = (
            "▷"
            if partner.name == self.connection.username and game.opponent.online
            else "*"
        )
        try:
            self.store.set(self.games[game]["ti"], 3, result)
        except KeyError:
            pass

        if (
            self.connection.stored_owner == self.connection.username
            and player.online
            and player.adjournment
        ):
            self._infobar_adjourned_message(game, player)
        elif not player.online and player in self.messages:
            self.messages[player].dismiss()
            # calling message.dismiss() might cause it to be removed from
            # self.messages in another callback, so we re-check
            if player in self.messages:
                del self.messages[player]

        return False

    def status_changed(self, player, prop, game):
        log.debug(f"AdjournedTabSection.status_changed: {repr(player)} {repr(game)}")
        try:
            message = self.messages[player]
        except KeyError:
            pass
        else:
            make_sensitive_if_available(message.buttons[0], player)

        self.onSelectionChanged(self.selection)
        return False

    def onAdjournedGameAdded(self, adm, game):
        if game not in self.games:
            partner = (
                game.bplayer
                if game.wplayer.name == game.opponent.name
                else game.wplayer
            )
            result = (
                "▷"
                if partner.name == self.connection.username and game.opponent.online
                else "*"
            )
            ti = self.store.append(
                [
                    game,
                    game.wplayer.name,
                    game.wrating,
                    result,
                    game.bplayer.name,
                    game.brating,
                    game.display_rated,
                    game.display_timecontrol,
                    game.game_type.display_text,
                    game.display_time,
                    game.sortable_time,
                ]
            )
            self.games[game] = {}
            self.games[game]["ti"] = ti
            self.games[game]["online_cid"] = game.opponent.connect(
                "notify::online", self.online_changed, game
            )
            self.games[game]["status_cid"] = game.opponent.connect(
                "notify::status", self.status_changed, game
            )

        if (
            self.connection.stored_owner == self.connection.username
            and game.opponent.online
        ):
            self._infobar_adjourned_message(game, game.opponent)

        return False

    def onHistoryGameAdded(self, adm, game):
        if game not in self.games:
            ti = self.store.append(
                [
                    game,
                    game.wplayer.name,
                    game.wrating,
                    reprResult[game.result],
                    game.bplayer.name,
                    game.brating,
                    game.display_rated,
                    game.display_timecontrol,
                    game.game_type.display_text,
                    game.display_time,
                    game.sortable_time,
                ]
            )
            self.games[game] = {}
            self.games[game]["ti"] = ti

        return False

    def onJournalGameAdded(self, adm, game):
        if game not in self.games:
            ti = self.store.append(
                [
                    game,
                    game.wplayer.name,
                    game.wrating,
                    reprResult[game.result],
                    game.bplayer.name,
                    game.brating,
                    game.display_rated,
                    game.display_timecontrol,
                    game.game_type.display_text,
                    game.display_time,
                    game.sortable_time,
                ]
            )
            self.games[game] = {}
            self.games[game]["ti"] = ti

        return False

    def onAdjournedGameRemoved(self, adm, game):
        if game in self.games:
            if self.store.iter_is_valid(self.games[game]["ti"]):
                self.store.remove(self.games[game]["ti"])
            if game.opponent.handler_is_connected(self.games[game]["online_cid"]):
                game.opponent.disconnect(self.games[game]["online_cid"])
            if game.opponent.handler_is_connected(self.games[game]["status_cid"]):
                game.opponent.disconnect(self.games[game]["status_cid"])
            if game.opponent in self.messages:
                self.messages[game.opponent].dismiss()
                if game.opponent in self.messages:
                    del self.messages[game.opponent]
            del self.games[game]

        return False

    def onHistoryGameRemoved(self, adm, game):
        if game in self.games:
            if self.store.iter_is_valid(self.games[game]["ti"]):
                self.store.remove(self.games[game]["ti"])
            del self.games[game]

        return False

    def onJournalGameRemoved(self, adm, game):
        if game in self.games:
            if self.store.iter_is_valid(self.games[game]["ti"]):
                self.store.remove(self.games[game]["ti"])
            del self.games[game]

        return False

    def onResignButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        self.connection.adm.resign(game)

    def onDrawButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        self.connection.adm.draw(game)

    def onAbortButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        self.connection.adm.abort(game)

    def onResumeButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        self.connection.adm.resume(game)

    def onPreviewButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        self.connection.adm.queryMoves(game)

    def onExamineButtonClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        game = model.get_value(sel_iter, 0)
        if self.connection.examined_game is None:
            self.connection.adm.examine(game)
        else:
            self.lounge.nonoWhileExamine(None)

    def onMygamesButtonClicked(self, button):
        self.connection.adm.queryAdjournments()
        self.connection.adm.queryHistory()
        self.connection.adm.queryJournal()

    def onGamePreview(self, adm, ficsgame):
        log.debug("Archived panel onGamePreview: %s" % ficsgame)

        timemodel = TimeModel(ficsgame.minutes * 60, ficsgame.inc)

        gamemodel = ICGameModel(self.connection, ficsgame, timemodel)

        # The players need to start listening for moves IN this method if they
        # want to be noticed of all moves the FICS server sends us from now on.
        # Hence the lazy loading is skipped.
        wplayer, bplayer = ficsgame.wplayer, ficsgame.bplayer
        player0tup = (
            REMOTE,
            ICPlayer,
            (
                gamemodel,
                wplayer.name,
                -1,
                WHITE,
                wplayer.long_name(),
                wplayer.getRatingByGameType(ficsgame.game_type),
            ),
            wplayer.long_name(),
        )
        player1tup = (
            REMOTE,
            ICPlayer,
            (
                gamemodel,
                bplayer.name,
                -1,
                BLACK,
                bplayer.long_name(),
                bplayer.getRatingByGameType(ficsgame.game_type),
            ),
            bplayer.long_name(),
        )

        perspective = perspective_manager.get_perspective("games")
        asyncio.create_task(
            perspective.generalStart(
                gamemodel,
                player0tup,
                player1tup,
                (StringIO(ficsgame.board.pgn), pgn, 0, -1),
            )
        )
        gamemodel.connect("game_started", self.on_game_start, ficsgame)

    def on_game_start(self, gamemodel, ficsgame):
        gamemodel.end(ficsgame.result, ficsgame.reason)



import asyncio
import math
import random
from os.path import basename
from urllib.request import urlopen
from urllib.parse import unquote

from gi.repository import Gtk, GObject, Pango

from pychess.Players.Human import Human
from pychess.Players.engineNest import discoverer
from pychess.System import uistuff, conf
from pychess.Utils.GameModel import GameModel
from pychess.Utils.IconLoader import get_pixbuf
from pychess.Utils.TimeModel import TimeModel
from pychess.Utils.const import (
    LOCAL,
    ARTIFICIAL,
    WHITE,
    BLACK,
    NORMALCHESS,
    LECTURE,
    LESSON,
    PUZZLE,
    ENDGAME,
)
from pychess.Variants import variants
from pychess.ic import ICLogon
from pychess.widgets import newGameDialog
from pychess.widgets.Background import giveBackground
from pychess.widgets.RecentChooser import recent_manager, recent_menu
from pychess.perspectives import perspective_manager
from pychess.perspectives.games import get_open_dialog
from pychess.perspectives.learn.LecturesPanel import LECTURES, start_lecture_from
from pychess.perspectives.learn.EndgamesPanel import ENDGAMES, start_endgame_from
from pychess.perspectives.learn.LessonsPanel import LESSONS, start_lesson_from
from pychess.perspectives.learn.PuzzlesPanel import PUZZLES, start_puzzle_from


class TaskerManager(Gtk.Table):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.border = 20
        giveBackground(self)
        self.connect("draw", self.expose)
        # self.set_homogeneous(True)

    def expose(self, widget, ctx):
        cairo_win = widget.get_window().cairo_create()

        for widget in self.widgets:
            x_loc = widget.get_allocation().x
            y_loc = widget.get_allocation().y
            width = widget.get_allocation().width
            height = widget.get_allocation().height

            cairo_win.move_to(x_loc - self.border, y_loc)
            cairo_win.curve_to(
                x_loc - self.border,
                y_loc - self.border / 2.0,
                x_loc - self.border / 2.0,
                y_loc - self.border,
                x_loc,
                y_loc - self.border,
            )
            cairo_win.line_to(x_loc + width, y_loc - self.border)
            cairo_win.curve_to(
                x_loc + width + self.border / 2.0,
                y_loc - self.border,
                x_loc + width + self.border,
                y_loc - self.border / 2.0,
                x_loc + width + self.border,
                y_loc,
            )
            cairo_win.line_to(x_loc + width + self.border, y_loc + height)
            cairo_win.curve_to(
                x_loc + width + self.border,
                y_loc + height + self.border / 2.0,
                x_loc + width + self.border / 2.0,
                y_loc + height + self.border,
                x_loc + width,
                y_loc + height + self.border,
            )
            cairo_win.line_to(x_loc, y_loc + height + self.border)
            cairo_win.curve_to(
                x_loc - self.border / 2.0,
                y_loc + height + self.border,
                x_loc - self.border,
                y_loc + height + self.border / 2.0,
                x_loc - self.border,
                y_loc + height,
            )

            style_ctxt = self.get_style_context()
            bgcolor = style_ctxt.lookup_color("p_bg_color")[1]
            darkcolor = style_ctxt.lookup_color("p_dark_color")[1]

            cairo_win.set_source_rgba(
                bgcolor.red, bgcolor.green, bgcolor.blue, bgcolor.alpha
            )
            cairo_win.fill()

            cairo_win.rectangle(
                x_loc - self.border, y_loc + height - 30, width + self.border * 2, 30
            )
            cairo_win.set_source_rgba(
                darkcolor.red, darkcolor.green, darkcolor.blue, darkcolor.alpha
            )
            cairo_win.fill()

    def calcSpacings(self, n):
        """Will yield ranges like
        ((.50,.50),)
        ((.66,.33), (.33,.66))
        ((.75,.25), (.50,.50), (.25,.75))
        ((.80,.20), (.60,.40), (.40,.60), (.20,.80))
        Used to create the centering in the table"""

        first = next = (n) / float(n + 1)
        for i in range(n):
            yield (next, 1 - next)
            next = first - (1 - next)

    def on_size_allocate(self, widget, allocation):
        window = self.get_window()
        if window is not None:
            window.invalidate_rect(self.get_allocation(), False)

    def packTaskers(self, *widgets):
        self.widgets = widgets

        for widget in widgets:
            widget.connect("size-allocate", self.on_size_allocate)
        root = math.sqrt(len(widgets))
        # Calculate number of rows
        rows = int(math.ceil(root))
        # Calculate number of filled out rows
        rrows = int(math.floor(root))
        # Calculate number of cols in filled out rows
        cols = int(math.ceil(len(widgets) / float(rows)))

        # Calculate spacings

        vspac = [s[0] for s in self.calcSpacings(rows)]
        hspac = [s[0] for s in self.calcSpacings(cols)]

        # Clear and set up new size

        for child in self.get_children():
            self.remove(child)

        self.props.n_columns = cols
        self.props.n_rows = rows

        # Add filled out rows

        for row in range(rows):
            for col in range(cols):
                widget = widgets[row * cols + col]
                alignment = Gtk.Alignment.new(hspac[col], vspac[row], 0, 0)
                alignment.add(widget)
                self.attach(alignment, col, col + 1, row, row + 1)
        return

        # Add last row

        if rows > rrows:
            lastrow = Gtk.HBox()
            # Calculate number of widgets in last row
            numw = len(widgets) - cols * rrows
            hspac = [s[0] for s in self.calcSpacings(numw)]
            for col, widget in enumerate(widgets[-numw:]):
                alignment = Gtk.Alignment.new(hspac[col], vspac[-1], 0, 0)
                alignment.add(widget)
                alignment.set_padding(
                    self.border, self.border, self.border, self.border
                )
                lastrow.pack_start(alignment, True, True, 0)

            self.attach(lastrow, 0, cols, rrows, rrows + 1)


tasker = TaskerManager()
tasker_widgets = uistuff.GladeWidgets("taskers.glade")


class NewGameTasker(Gtk.Alignment):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.widgets = widgets = tasker_widgets
        tasker = widgets["newGameTasker"]
        tasker.unparent()
        self.add(tasker)

        startButton = self.widgets["startButton"]
        startButton.set_name("startButton")
        combo = Gtk.ComboBox()
        uistuff.createCombo(
            combo,
            [
                (get_pixbuf("glade/white.png"), _("White")),
                (get_pixbuf("glade/black.png"), _("Black")),
                (get_pixbuf("glade/random.png"), _("Random")),
            ],
        )
        widgets["colorDock"].add(combo)
        if combo.get_active() < 0:
            combo.set_active(0)
        widgets["yourColorLabel"].set_mnemonic_widget(combo)

        # We need to wait until after engines have been discovered, to init the
        # playerCombos. We use connect_after to make sure, that newGameDialog
        # has also had time to init the constants we share with them.
        self.playerCombo = Gtk.ComboBox()
        widgets["opponentDock"].add(self.playerCombo)
        discoverer.connect_after(
            "all_engines_discovered", self.__initPlayerCombo, widgets
        )
        widgets["opponentLabel"].set_mnemonic_widget(self.playerCombo)

        def on_skill_changed(scale):
            # Just to make sphinx happy...
            try:
                pix = newGameDialog.skillToIconLarge[int(scale.get_value())]
                widgets["skillImage"].set_from_pixbuf(pix)
            except TypeError:
                pass

        widgets["skillSlider"].connect("value-changed", on_skill_changed)
        on_skill_changed(widgets["skillSlider"])

        widgets["startButton"].connect("clicked", self.startClicked)
        self.widgets["opendialog1"].connect("clicked", self.openDialogClicked)

    def __initPlayerCombo(self, discoverer, widgets):
        combo = self.playerCombo
        uistuff.createCombo(combo, newGameDialog.playerItems[0])
        if combo.get_active() < 0:
            combo.set_active(1)
            uistuff.keep(self.playerCombo, "newgametasker_playercombo")

            def on_playerCombobox_changed(widget):
                widgets["skillSlider"].props.visible = widget.get_active() > 0

            combo.connect("changed", on_playerCombobox_changed)

            uistuff.keep(widgets["skillSlider"], "taskerSkillSlider")
            widgets["skillSlider"].set_no_show_all(True)
            on_playerCombobox_changed(self.playerCombo)

    def openDialogClicked(self, button):
        newGameDialog.NewGameMode.run()

    def startClicked(self, button):
        color = self.widgets["colorDock"].get_child().get_active()
        if color == 2:
            color = random.choice([WHITE, BLACK])

        opp = self.widgets["opponentDock"].get_child()
        tree_iter = opp.get_active_iter()
        if tree_iter is not None:
            model = opp.get_model()
            engine = model[tree_iter][1]

        opponent = self.widgets["opponentDock"].get_child().get_active()
        difficulty = int(self.widgets["skillSlider"].get_value())

        gamemodel = GameModel(TimeModel(5 * 60, 0))

        name = conf.get("firstName")
        player0tup = (LOCAL, Human, (color, name), name)
        if opponent == 0:
            name = conf.get("secondName")
            player1tup = (LOCAL, Human, (1 - color, name), name)
        else:
            engine = discoverer.getEngineByName(engine)
            name = discoverer.getName(engine)
            player1tup = (
                ARTIFICIAL,
                discoverer.initPlayerEngine,
                (engine, 1 - color, difficulty, variants[NORMALCHESS], 5 * 60, 0),
                name,
            )

        perspective = perspective_manager.get_perspective("games")
        if color == WHITE:
            asyncio.create_task(
                perspective.generalStart(gamemodel, player0tup, player1tup)
            )
        else:
            asyncio.create_task(
                perspective.generalStart(gamemodel, player1tup, player0tup)
            )


class InternetGameTasker(Gtk.Alignment):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.widgets = tasker_widgets
        tasker = self.widgets["internetGameTasker"]
        tasker.unparent()
        self.add(tasker)

        if ICLogon.dialog is None:
            ICLogon.dialog = ICLogon.ICLogon()

        liststore = Gtk.ListStore(str, bool)
        liststore.append(["FICS", True])
        liststore.append(["ICC", False])
        self.ics_combo = self.widgets["ics_combo"]
        self.ics_combo.set_model(liststore)
        renderer_text = Gtk.CellRendererText()
        self.ics_combo.pack_start(renderer_text, True)
        self.ics_combo.add_attribute(renderer_text, "text", 0)

        # ICC doesn't support telnet connection any more
        # See https://github.com/pychess/pychess/issues/2315

        def cell_data_func(combo, cell, model, iter_, data):
            """Customize cell rendering based on the 'enabled' column."""
            is_enabled = model[iter_][1]
            cell.set_property("sensitive", is_enabled)  # Gray out if disabled

        self.ics_combo.set_cell_data_func(renderer_text, cell_data_func, None)

        self.ics_combo.connect("changed", ICLogon.dialog.on_ics_combo_changed)

        # self.ics_combo.set_active(conf.get("ics_combo"))
        self.ics_combo.set_active(0)  # select FICS

        self.widgets["connectButton"].connect("clicked", self.connectClicked)
        self.widgets["opendialog2"].connect("clicked", self.openDialogClicked)

        big_start = get_pixbuf("glade/internet.png")
        self.widgets["startIcon"].set_from_pixbuf(big_start)

        uistuff.keep(self.widgets["ics_combo"], "ics_combo")
        uistuff.keep(self.widgets["autoLogin"], "autoLogin")

    def openDialogClicked(self, button):
        ICLogon.run()

    def connectClicked(self, button):
        ICLogon.run()
        if not ICLogon.dialog.connection:
            ICLogon.dialog.widgets["connectButton"].clicked()


class LearnTasker(Gtk.Alignment):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.widgets = tasker_widgets
        tasker = self.widgets["learnTasker"]
        tasker.unparent()
        self.add(tasker)

        startButton = self.widgets["learnButton"]
        startButton.set_name("learnButton")

        categorystore = Gtk.ListStore(int, str)

        learn_mapping = {
            LECTURE: (_("Lectures"), LECTURES),
            LESSON: (_("Lessons"), LESSONS),
            PUZZLE: (_("Puzzles"), PUZZLES),
            ENDGAME: (_("Endgames"), ENDGAMES),
        }
        for key, value in learn_mapping.items():
            categorystore.append([key, value[0]])

        self.category_combo = self.widgets["category_combo"]
        self.category_combo.set_model(categorystore)
        renderer = Gtk.CellRendererText()
        self.category_combo.pack_start(renderer, True)
        self.category_combo.add_attribute(renderer, "text", 1)

        self.learnstore = Gtk.ListStore(str, str)
        self.learn_combo = self.widgets["learn_combo"]
        self.learn_combo.set_model(self.learnstore)
        renderer_text = Gtk.CellRendererText()
        renderer_text.set_property("width-chars", 30)
        renderer_text.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.learn_combo.pack_start(renderer_text, True)
        self.learn_combo.add_attribute(renderer_text, "text", 1)
        self.learn_combo.set_active(0)

        def on_category_changed(combo):
            tree_iter = combo.get_active_iter()
            if tree_iter is None:
                return
            else:
                model = combo.get_model()
                self.category = model[tree_iter][0]

                self.learnstore.clear()
                if self.category == LECTURE:
                    for file_name, title, author in LECTURES:
                        self.learnstore.append([file_name, title])
                elif self.category == LESSON:
                    for file_name, title, author in LESSONS:
                        self.learnstore.append([file_name, title])
                elif self.category == PUZZLE:
                    for file_name, title, author in PUZZLES:
                        self.learnstore.append([file_name, title])
                elif self.category == ENDGAME:
                    for pieces, title in ENDGAMES:
                        self.learnstore.append([pieces, title])

                learn = conf.get("learncombo%s" % self.category)
                self.learn_combo.set_active(learn)

                def on_learn_changed(combo):
                    tree_iter = combo.get_active_iter()
                    if tree_iter is None:
                        return
                    else:
                        model = combo.get_model()
                        newlearn = model.get_path(tree_iter)[0]
                        conf.set("learncombo%s" % self.category, newlearn)

                self.learn_combo.connect("changed", on_learn_changed)

        self.category_combo.connect("changed", on_category_changed)
        self.category = conf.get("categorycombo")
        self.category_combo.set_active(self.category)

        dictionary = get_pixbuf("glade/dictionary.png")
        self.widgets["dictionaryIcon"].set_from_pixbuf(dictionary)

        uistuff.keep(self.widgets["category_combo"], "categorycombo")

        self.widgets["opendialog4"].connect("clicked", self.openDialogClicked)
        self.widgets["learnButton"].connect("clicked", self.learnClicked)

    def openDialogClicked(self, button):
        perspective = perspective_manager.get_perspective("learn")
        perspective.activate()

    def learnClicked(self, button):
        perspective = perspective_manager.get_perspective("learn")
        perspective.activate()

        tree_iter = self.learn_combo.get_active_iter()
        if tree_iter is None:
            return
        else:
            model = self.learn_combo.get_model()
            source = model[tree_iter][0]

        if self.category == LECTURE:
            start_lecture_from(source)
        elif self.category == LESSON:
            start_lesson_from(source)
        elif self.category == PUZZLE:
            start_puzzle_from(source)
        elif self.category == ENDGAME:
            start_endgame_from(source)


class DatabaseTasker(Gtk.Alignment):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.widgets = tasker_widgets
        tasker = self.widgets["databaseTasker"]
        tasker.unparent()
        self.add(tasker)

        startButton = self.widgets["openButton"]
        startButton.set_name("openButton")

        liststore = Gtk.ListStore(str, str)

        self.recent_combo = self.widgets["recent_combo"]
        self.recent_combo.set_model(liststore)
        renderer_text = Gtk.CellRendererText()
        renderer_text.set_property("width-chars", 30)
        renderer_text.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.recent_combo.pack_start(renderer_text, True)
        self.recent_combo.add_attribute(renderer_text, "text", 1)

        self.on_recent_menu_changed(recent_manager, liststore)
        recent_manager.connect("changed", self.on_recent_menu_changed, liststore)

        openimage = get_pixbuf("glade/open48.png")
        self.widgets["openImage"].set_from_pixbuf(openimage)

        self.widgets["opendialog3"].connect("clicked", self.openDialogClicked)
        self.widgets["openButton"].connect("clicked", self.openClicked)

    def on_recent_menu_changed(self, manager, liststore):
        liststore.clear()
        # Just to make sphinx happy...
        try:
            for uri in recent_menu.get_uris():
                liststore.append(
                    (
                        uri,
                        basename(unquote(uri)),
                    )
                )
        except TypeError:
            pass
        self.recent_combo.set_active(0)

    def openDialogClicked(self, button):
        dialog = get_open_dialog()

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filenames = dialog.get_filenames()
        else:
            filenames = None

        dialog.destroy()

        if filenames is not None:
            for filename in filenames:
                if filename.lower().endswith(".fen"):
                    newGameDialog.loadFileAndRun(filename)
                else:
                    perspective = perspective_manager.get_perspective("database")
                    perspective.open_chessfile(filename)

    def openClicked(self, button):
        if self.widgets["createNew"].get_active():
            perspective = perspective_manager.get_perspective("database")
            perspective.create_database()

        else:
            tree_iter = self.recent_combo.get_active_iter()
            if tree_iter is None:
                return
            else:
                model = self.recent_combo.get_model()
                uri = model[tree_iter][0]

            try:
                urlopen(unquote(uri)).close()
                perspective = perspective_manager.get_perspective("database")
                perspective.open_chessfile(unquote(uri))
                recent_manager.add_item(uri)
            except OSError:
                # shomething wrong whit the uri
                recent_manager.remove_item(uri)


new_game_tasker, internet_game_tasker, database_tasker, learn_tasker = (
    NewGameTasker(),
    InternetGameTasker(),
    DatabaseTasker(),
    LearnTasker(),
)
tasker.packTaskers(new_game_tasker, database_tasker, internet_game_tasker, learn_tasker)



"""
This module differs from leval in that it is not optimized for speed.
It checks differences between last and current board, and returns not
scores, but strings describing the differences.
Can be used for commenting on board changes.
"""

from .ldata import (
    brank48,
    brank67,
    bitPosArray,
    left,
    right,
    stonewall,
    distance,
    isolaniMask,
    fileBits,
    passedScores,
    passedPawnMask,
    fromToRay,
    outpost,
    FILE,
    PIECE_VALUES,
    ray45,
    ray135,
    ray90,
    ray00,
)

from .bitboard import clearBit, lastBit, iterBits
from pychess.Utils.lutils.attack import staticExchangeEvaluate, getAttacks, defends
from pychess.Utils.lutils.lmove import toSAN, TCORD, FCORD, FLAG, PROMOTE_PIECE
from pychess.Utils.const import (
    BLACK,
    WHITEWON,
    BLACKWON,
    DRAW,
    reprFile,
    reprCord,
    QUEEN_CASTLE,
    KING_CASTLE,
    WHITE,
    KNIGHT,
    BISHOP,
    ROOK,
    PAWN,
    KING,
    QUEEN,
    EMPTY,
    PROMOTIONS,
    FISCHERRANDOMCHESS,
    H7,
    G6,
    A7,
    B6,
    H2,
    G3,
    A2,
    B3,
    B7,
    G7,
    B2,
    G2,
    B_OO,
    B_OOO,
    W_OO,
    W_OOO,
)
from pychess.Utils.lutils.lmovegen import genCaptures, genAllMoves, newMove
from pychess.Utils.lutils.validator import validateMove
from pychess.Utils.repr import reprColor, reprPiece
from . import leval


def join(items):
    if len(items) == 1:
        return items[0]
    else:
        s = "{} {} {}".format(items[-2], _("and"), items[-1])
        if len(items) > 2:
            s = ", ".join(items[:-2] + [s])
        return s


# Functions can be of types:
#   * Final: Will be shown alone: "mates", "draws"
#   * Moves (s): Will always be shown: "put into *"
#   * Prefix: Will always be shown: "castles", "promotes"
#   * Attack: Will always be shown: "threaten", "preassures", "defendes"
#   * Simple: (s) Max one will be shown: "develops", "activity"
#   * State: (s) Will always be shown: "new *"
#   * Tip: (s) Will sometimes be shown: "pawn storm", "cramped position"


def final_status(model, ply, phase):
    if ply == model.ply:
        if model.status == DRAW:
            yield _("draws")
        elif model.status in (WHITEWON, BLACKWON):
            yield _("mates")


def offencive_moves_check(model, ply, phase):
    if model.getBoardAtPly(ply).board.isChecked():
        yield _("puts opponent in check")


def defencive_moves_safety(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board

    if board.arBoard[TCORD(model.getMoveAtPly(ply - 1).move)] != KING:
        return

    color = oldboard.color
    opcolor = 1 - color

    delta_eval_king = leval.evalKing(board, color, phase) - leval.evalKing(
        oldboard, color, phase
    )

    # PyChess points tropism to queen for phase <= 3. Thus we set a high phase
    delta_eval_tropism = leval.evalKingTropism(
        board, opcolor, 10
    ) - leval.evalKingTropism(oldboard, opcolor, 10)

    # Notice, that tropism was negative
    delta_score = delta_eval_king - delta_eval_tropism / 2

    if delta_score > 35:
        yield _("improves king safety")
    elif delta_score > 15:
        yield _("slightly improves king safety")


def offencive_moves_rook(model, ply, phase):
    move = model.getMoveAtPly(ply - 1).move
    tcord = TCORD(move)
    board = model.getBoardAtPly(ply).board
    color = 1 - board.color
    opcolor = 1 - color

    # We also detect rook-to-open castlings
    if board.arBoard[tcord] == KING:
        if FLAG(move) == QUEEN_CASTLE:
            tcord = tcord + 1
        elif FLAG(move) == KING_CASTLE:
            tcord = tcord - 1

    if board.arBoard[tcord] != ROOK:
        return

    color = 1 - board.color
    opcolor = 1 - color

    pawns = board.boards[color][PAWN]
    oppawns = board.boards[opcolor][PAWN]

    ffile = fileBits[FILE(FCORD(move))]
    tfile = fileBits[FILE(tcord)]

    if ffile & pawns and not tfile & pawns and bin(pawns).count("1") >= 3:
        if not tfile & oppawns:
            yield _("moves a rook to an open file")
        else:
            yield _("moves a rook to a half-open file")


def offencive_moves_fianchetto(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    tcord = TCORD(model.getMoveAtPly(ply - 1).move)
    movingcolor = 1 - board.color

    if movingcolor == WHITE:
        if board.castling & W_OO and tcord == G2:
            yield _("moves bishop into fianchetto: %s") % "g2"
        elif board.castling & W_OOO and tcord == B2:
            yield _("moves bishop into fianchetto: %s") % "b2"
    else:
        if board.castling & B_OO and tcord == G7:
            yield _("moves bishop into fianchetto: %s") % "g7"
        elif board.castling & B_OOO and tcord == B7:
            yield _("moves bishop into fianchetto: %s") % "b7"


def prefix_type(model, ply, phase):
    flag = FLAG(model.getMoveAtPly(ply - 1).move)

    if flag in PROMOTIONS:
        yield _("promotes a Pawn to a %s") % reprPiece[PROMOTE_PIECE(flag)]

    elif flag in (KING_CASTLE, QUEEN_CASTLE):
        yield _("castles")


def attack_type(model, ply, phase):
    # We set bishop value down to knight value, as it is what most people expect
    bishopBackup = PIECE_VALUES[BISHOP]
    PIECE_VALUES[BISHOP] = PIECE_VALUES[KNIGHT]

    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board

    if ply - model.lowply >= 2:
        oldmove = model.getMoveAtPly(ply - 2).move
        oldboard3 = model.getBoardAtPly(ply - 2).board
    else:
        oldmove = None
    move = model.getMoveAtPly(ply - 1).move
    tcord = TCORD(move)

    if oldboard.arBoard[tcord] != EMPTY:
        if not (
            board.variant == FISCHERRANDOMCHESS
            and FLAG(move) in (KING_CASTLE, QUEEN_CASTLE)
        ):
            if (
                oldmove
                and oldboard3.arBoard[TCORD(oldmove)] != EMPTY
                and TCORD(oldmove) == tcord
            ):
                yield _("takes back material")
            else:
                see = staticExchangeEvaluate(oldboard, move)
                if see < 0:
                    yield _("sacrifices material")
                elif see == 0:
                    yield _("exchanges material")
                elif see > 0:
                    yield _("captures material")
    else:
        see = staticExchangeEvaluate(oldboard, move)
        if see < 0:
            yield _("sacrifices material")

    PIECE_VALUES[BISHOP] = bishopBackup


def defencive_moves_tactic(model, ply, phase):
    # ------------------------------------------------------------------------ #
    # Test if we threat something, or at least put more pressure on it         #
    # ------------------------------------------------------------------------ #

    # We set bishop value down to knight value, as it is what most people expect
    bishopBackup = PIECE_VALUES[BISHOP]
    PIECE_VALUES[BISHOP] = PIECE_VALUES[KNIGHT]

    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board
    move = model.getMoveAtPly(ply - 1).move
    fcord = FCORD(move)
    tcord = TCORD(move)

    found_threatens = []
    found_increases = []

    # What do we attack now?
    board.setColor(1 - board.color)
    for ncap in genCaptures(board):
        # getCaptures also generate promotions
        if FLAG(ncap) in PROMOTIONS:
            continue

        # We are only interested in the attacks of the piece we just moved
        if FCORD(ncap) != TCORD(move):
            continue

        # We don't want to move back
        if TCORD(ncap) == FCORD(move):
            continue

        # We don't thread the king. We check him! (in another function)
        if board.arBoard[TCORD(ncap)] == KING:
            continue

        # If we also was able to attack that cord last time, we don't care
        if validateMove(oldboard, newMove(FCORD(move), TCORD(ncap))):
            continue

        # Test if we threats our enemy, at least more than before
        see0 = staticExchangeEvaluate(oldboard, TCORD(ncap), 1 - oldboard.color)
        see1 = staticExchangeEvaluate(board, TCORD(ncap), 1 - oldboard.color)
        if see1 > see0:
            # If a new winning capture has been created
            if see1 > 0:
                # Find the easiest attack
                attacks = getAttacks(board, TCORD(ncap), board.color)
                v, cord = min(
                    (PIECE_VALUES[board.arBoard[fc]], fc) for fc in iterBits(attacks)
                )
                easiestAttack = newMove(cord, TCORD(ncap))
                found_threatens.append(toSAN(board, easiestAttack, True))

            # Even though we might not yet be strong enough, we might still
            # have strengthened another friendly attack
            else:
                found_increases.append(reprCord[TCORD(ncap)])

    board.setColor(1 - board.color)

    # -------------------------------------------------------------------- #
    # Test if we defend a one of our pieces                                #
    # -------------------------------------------------------------------- #

    found_defends = []

    # Test which pieces were under attack
    used = []
    for ncap in genCaptures(board):
        # getCaptures also generate promotions
        if FLAG(ncap) in PROMOTIONS:
            continue

        # We don't want to know about the same cord more than once
        if TCORD(ncap) in used:
            continue
        used.append(TCORD(ncap))

        # If the attack was poining on the piece we just moved, we ignore it
        if TCORD(ncap) == FCORD(move) or TCORD(ncap) == TCORD(move):
            continue

        # If we were already defending the piece, we don't send a new
        # message
        if defends(oldboard, FCORD(move), TCORD(ncap)):
            continue

        # If the attack was not strong, we ignore it
        see = staticExchangeEvaluate(oldboard, ncap)
        if see < 0:
            continue

        v = defends(board, TCORD(move), TCORD(ncap))

        # If the defend didn't help, it doesn't matter. Like defending a
        # bishop, threatened by a pawn, with a queen.
        # But on the other hand - it might still be a defend...
        # newsee = staticExchangeEvaluate(board, ncap)
        # if newsee <= see: continue

        if v:
            found_defends.append(reprCord[TCORD(ncap)])

    # ------------------------------------------------------------------------ #
    # Test if we are rescuing an otherwise exposed piece                       #
    # ------------------------------------------------------------------------ #

    # Rescuing is only an option, if our own move wasn't an attack
    if oldboard.arBoard[tcord] == EMPTY:
        see0 = staticExchangeEvaluate(oldboard, fcord, oldboard.color)
        see1 = staticExchangeEvaluate(board, tcord, oldboard.color)
        if see1 > see0 and see1 > 0:
            yield _("rescues a %s") % reprPiece[board.arBoard[tcord]].lower()

    if found_threatens:
        yield _("threatens to win material by %s") % join(found_threatens)
    if found_increases:
        yield _("increases the pressure on %s") % join(found_increases)
    if found_defends:
        yield _("defends %s") % join(found_defends)

    PIECE_VALUES[BISHOP] = bishopBackup


def offencive_moves_pin(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    move = model.getMoveAtPly(ply - 1).move
    fcord = FCORD(move)
    tcord = TCORD(move)
    piece = board.arBoard[tcord]

    ray = 0
    if piece in (BISHOP, QUEEN):
        ray |= (ray45[tcord] | ray135[tcord]) & ~(ray45[fcord] | ray135[fcord])
    if piece in (ROOK, QUEEN):
        ray |= (ray00[tcord] | ray90[tcord]) & ~(ray00[fcord] | ray90[fcord])

    if ray:
        for c in iterBits(ray & board.friends[board.color]):
            # We don't pin on pieces that are less worth than us
            if not PIECE_VALUES[piece] < PIECE_VALUES[board.arBoard[c]]:
                continue
            # There should be zero friendly pieces in between
            ray = fromToRay[tcord][c]
            if ray & board.friends[1 - board.color]:
                continue
            # There should be exactly one opponent piece in between
            op = clearBit(ray & board.friends[board.color], c)
            if bin(op).count("1") != 1:
                continue
            # The king can't be pinned
            pinned = lastBit(op)
            oppiece = board.arBoard[pinned]
            if oppiece == KING:
                continue
            # Yield
            yield _("pins an enemy %(oppiece)s on the %(piece)s at %(cord)s") % {
                "oppiece": reprPiece[oppiece].lower(),
                "piece": reprPiece[board.arBoard[c]].lower(),
                "cord": reprCord[c],
            }


def state_outpost(model, ply, phase):
    if phase >= 6:
        # Doesn't make sense in endgame
        return

    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board

    wpawns = board.boards[WHITE][PAWN]
    oldwpawns = oldboard.boards[WHITE][PAWN]
    bpawns = board.boards[BLACK][PAWN]
    oldbpawns = oldboard.boards[BLACK][PAWN]

    wpieces = board.boards[WHITE][BISHOP] | board.boards[WHITE][KNIGHT]
    oldwpieces = oldboard.boards[WHITE][BISHOP] | oldboard.boards[WHITE][KNIGHT]
    bpieces = board.boards[BLACK][BISHOP] | board.boards[BLACK][KNIGHT]
    oldbpieces = oldboard.boards[BLACK][BISHOP] | oldboard.boards[BLACK][KNIGHT]

    for cord in iterBits(wpieces):
        sides = isolaniMask[FILE(cord)]
        front = passedPawnMask[WHITE][cord]
        if (
            outpost[WHITE][cord]
            and not bpawns & sides & front
            and (not oldwpieces & bitPosArray[cord] or oldbpawns & sides & front)
        ):
            yield 35, _("White has a new piece in outpost: %s") % reprCord[cord]

    for cord in iterBits(bpieces):
        sides = isolaniMask[FILE(cord)]
        front = passedPawnMask[BLACK][cord]
        if (
            outpost[BLACK][cord]
            and not wpawns & sides & front
            and (not oldbpieces & bitPosArray[cord] or oldwpawns & sides & front)
        ):
            yield 35, _("Black has a new piece in outpost: %s") % reprCord[cord]


def state_pawn(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board
    color = 1 - board.color
    opcolor = 1 - color

    move = model.getMoveAtPly(ply - 1).move

    pawns = board.boards[color][PAWN]
    oppawns = board.boards[opcolor][PAWN]
    oldpawns = oldboard.boards[color][PAWN]
    oldoppawns = oldboard.boards[opcolor][PAWN]

    # Passed pawns
    for cord in iterBits(pawns):
        if not oppawns & passedPawnMask[color][cord]:
            if color == WHITE:
                frontCords = fromToRay[cord][cord | 56]
            else:
                frontCords = fromToRay[cord][cord & 7]
            if frontCords & pawns:
                continue

            # Was this a passed pawn before?
            if (
                oldpawns & bitPosArray[cord]
                and not oldoppawns & passedPawnMask[color][cord]
                and not frontCords & oldpawns
            ):
                continue

            # Is this just a passed pawn that has been moved?
            if TCORD(move) == cord:
                frontCords |= bitPosArray[cord]
                if (
                    not frontCords & oldpawns
                    and not oldoppawns & passedPawnMask[color][FCORD(move)]
                ):
                    continue

            score = passedScores[color][cord >> 3] * phase
            yield score, _("%(color)s has a new passed pawn on %(cord)s") % {
                "color": reprColor[color],
                "cord": reprCord[cord],
            }

    # Double pawns
    found_doubles = []
    found_halfopen_doubles = []
    found_white_isolates = []
    found_black_isolates = []
    for file in range(8):
        bits = fileBits[file]

        count = bin(pawns & bits).count("1")
        oldcount = bin(oldpawns & bits).count("1")
        opcount = bin(oppawns & bits).count("1")
        oldopcount = bin(oldoppawns & bits).count("1")

        # Single pawn -> double pawns
        if count > oldcount >= 1:
            if not opcount:
                found_halfopen_doubles.append(reprFile[file])
            else:
                found_doubles.append(reprFile[file])

        # Closed file double pawn -> half-open file double pawn
        elif count > 1 and opcount == 0 and oldopcount > 0:
            found_halfopen_doubles.append(reprFile[file])

        # Isolated pawns
        if color == WHITE:
            wpawns = pawns
            oldwpawns = oldpawns
            bpawns = oppawns
            oldbpawns = oldoppawns
        else:
            bpawns = pawns
            oldbpawns = oldpawns
            wpawns = oppawns
            oldwpawns = oldoppawns

        if (
            wpawns & bits
            and not wpawns & isolaniMask[file]
            and (not oldwpawns & bits or oldwpawns & isolaniMask[file])
        ):
            found_white_isolates.append(reprFile[file])

        if (
            bpawns & bits
            and not bpawns & isolaniMask[file]
            and (not oldbpawns & bits or oldbpawns & isolaniMask[file])
        ):
            found_black_isolates.append(reprFile[file])

    # We need to take care of 'worstcases' like: "got new double pawns in the a
    # file, in the half-open b, c and d files and in the open e and f files"

    doubles_count = len(found_doubles) + len(found_halfopen_doubles)
    if doubles_count > 0:
        parts = []
        for type_, list_ in (
            ("", found_doubles),
            (_("half-open") + " ", found_halfopen_doubles),
        ):
            if len(list_) == 1:
                parts.append(_("in the %(x)s%(y)s file") % {"x": type_, "y": list_[0]})
            elif len(list_) >= 2:
                parts.append(
                    _("in the %(x)s%(y)s files") % {"x": type_, "y": join(list_)}
                )

        if doubles_count == 1:
            s = _("%(color)s got a double pawn %(place)s")
        else:
            s = _("%(color)s got new double pawns %(place)s")

        yield (8 + phase) * 2 * doubles_count, s % {
            "color": reprColor[color],
            "place": join(parts),
        }

    for color_, list_ in ((WHITE, found_white_isolates), (BLACK, found_black_isolates)):
        if list_:
            yield 20 * len(list_), ngettext(
                "%(color)s got an isolated pawn in the %(x)s file",
                "%(color)s got isolated pawns in the %(x)s files",
                len(list_),
            ) % {"color": reprColor[color_], "x": join(list_)}

    # Stone wall
    if (
        stonewall[color] & pawns == stonewall[color]
        and stonewall[color] & oldpawns != stonewall[color]
    ):
        yield 10, _("%s moves pawns into stonewall formation") % reprColor[color]


def state_destroysCastling(model, ply, phase):
    """Does the move destroy the castling ability of the opponent"""

    # If the move is a castling, nobody will every care if the castling
    # possibilities has changed
    if FLAG(model.getMoveAtPly(ply - 1).move) in (QUEEN_CASTLE, KING_CASTLE):
        return

    oldcastling = model.getBoardAtPly(ply - 1).board.castling
    castling = model.getBoardAtPly(ply).board.castling

    if oldcastling & W_OOO and not castling & W_OOO:
        if oldcastling & W_OO and not castling & W_OO:
            yield 900 / phase, _("%s can no longer castle") % reprColor[WHITE]
        else:
            yield 400 / phase, _("%s can no longer castle in queenside") % reprColor[
                WHITE
            ]
    elif oldcastling & W_OO and not castling & W_OO:
        yield 500 / phase, _("%s can no longer castle in kingside") % reprColor[WHITE]

    if oldcastling & B_OOO and not castling & B_OOO:
        if oldcastling & B_OO and not castling & B_OO:
            yield 900 / phase, _("%s can no longer castle") % reprColor[BLACK]
        else:
            yield 400 / phase, _("%s can no longer castle in queenside") % reprColor[
                BLACK
            ]
    elif oldcastling & B_OO and not castling & B_OO:
        yield 500 / phase, _("%s can no longer castle in kingside") % reprColor[BLACK]


def state_trappedBishops(model, ply, phase):
    """Check for bishops trapped at A2/H2/A7/H7"""

    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board

    opcolor = board.color

    move = model.getMoveAtPly(ply - 1).move
    tcord = TCORD(move)

    # Only a pawn is able to trap a bishop
    if board.arBoard[tcord] != PAWN:
        return

    if tcord == B3:
        cord = A2
    elif tcord == G3:
        cord = H2
    elif tcord == B6:
        cord = A7
    elif tcord == G6:
        cord = H7
    else:
        return

    s = leval.evalTrappedBishops(board, opcolor)
    olds = leval.evalTrappedBishops(oldboard, opcolor)

    # We have got more points -> We have trapped a bishop
    if s > olds:
        yield 300 / phase, _("%(opcolor)s has a new trapped bishop on %(cord)s") % {
            "opcolor": reprColor[opcolor],
            "cord": reprCord[cord],
        }


def simple_tropism(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board
    color = oldboard.color

    move = model.getMoveAtPly(ply - 1).move
    fcord = FCORD(move)
    tcord = TCORD(move)
    arBoard = board.arBoard

    if arBoard[tcord] != PAWN:
        score = leval.evalKingTropism(board, color, phase)
        oldscore = leval.evalKingTropism(oldboard, color, phase)
    else:
        if color == WHITE:
            rank23 = brank67[BLACK]
        else:
            rank23 = brank67[WHITE]
        if bitPosArray[fcord] & rank23:
            yield 2, _("develops a pawn: %s") % reprCord[tcord]
        else:
            yield 1, _("brings a pawn closer to the backrow: %s") % reprCord[tcord]
        return

    king = board.kings[color]
    opking = board.kings[1 - color]

    if score > oldscore:
        # in FISCHERRANDOMCHESS unusual casting case the tcord is
        # the involved rook's position, not the king's destination!
        flag = move >> 12
        if flag in (KING_CASTLE, QUEEN_CASTLE):
            piece = KING
        else:
            piece = arBoard[tcord]
        if phase >= 5 or distance[piece][fcord][opking] < distance[piece][fcord][king]:
            yield score - oldscore, _(
                "brings a %(piece)s closer to enemy king: %(cord)s"
            ) % {"piece": reprPiece[piece], "cord": reprCord[tcord]}
        else:
            yield (score - oldscore) * 2, _("develops a %(piece)s: %(cord)s") % {
                "piece": reprPiece[piece].lower(),
                "cord": reprCord[tcord],
            }


def simple_activity(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    oldboard = model.getBoardAtPly(ply - 1).board
    move = model.getMoveAtPly(ply - 1).move
    fcord = FCORD(move)
    tcord = TCORD(move)

    board.setColor(1 - board.color)
    moves = len([m for m in genAllMoves(board) if FCORD(m) == tcord])
    board.setColor(1 - board.color)
    oldmoves = len([m for m in genAllMoves(oldboard) if FCORD(m) == fcord])

    if moves > oldmoves:
        yield (moves - oldmoves) / 2, _("places a %(piece)s more active: %(cord)s") % {
            "piece": reprPiece[board.arBoard[tcord]].lower(),
            "cord": reprCord[tcord],
        }


def tip_pawnStorm(model, ply, phase):
    """If players are castled in different directions we should storm in
    opponent side"""

    if phase >= 6:
        # We don't use this in endgame
        return

    board = model.getBoardAtPly(ply).board

    # if not board.hasCastled[WHITE] or not board.hasCastled[BLACK]:
    #    # Only applies after castling for both sides
    #    return

    wking = board.boards[WHITE][KING]
    bking = board.boards[BLACK][KING]
    wleft = bin(board.boards[WHITE][PAWN] & left).count("1")
    wright = bin(board.boards[WHITE][PAWN] & right).count("1")
    bleft = bin(board.boards[BLACK][PAWN] & left).count("1")
    bright = bin(board.boards[BLACK][PAWN] & right).count("1")

    if wking & left and bking & right:
        if wright > bright:
            yield (wright + 3 - bright) * 10, _("White should do pawn storm in right")
        elif bleft > wleft:
            yield (bright + 3 - wright) * 10, _("Black should do pawn storm in left")
    if wking & right and bking & left:
        if wleft > bleft:
            yield (wleft + 3 - bleft) * 10, _("White should do pawn storm in left")
        if bright > wright:
            yield (bleft + 3 - wleft) * 10, _("Black should do pawn storm in right")


def tip_mobility(model, ply, phase):
    board = model.getBoardAtPly(ply).board
    colorBackup = board.color

    # People need a chance to get developed
    # if model.ply < 16:
    #    return

    board.setColor(WHITE)
    wmoves = len(
        [
            move
            for move in genAllMoves(board)
            if KNIGHT <= board.arBoard[FCORD(move)] <= QUEEN
            and bitPosArray[TCORD(move)] & brank48[WHITE]
            and staticExchangeEvaluate(board, move) >= 0
        ]
    )

    board.setColor(BLACK)
    bmoves = len(
        [
            move
            for move in genAllMoves(board)
            if KNIGHT <= board.arBoard[FCORD(move)] <= QUEEN
            and bitPosArray[TCORD(move)] & brank48[BLACK]
            and staticExchangeEvaluate(board, move) >= 0
        ]
    )

    board.setColor(colorBackup)

    if wmoves - phase >= (bmoves + 1) * 7:
        yield wmoves - bmoves, _("Black has a rather cramped position")
    elif wmoves - phase >= (bmoves + 1) * 3:
        yield wmoves - bmoves, _("Black has a slightly cramped position")
    elif bmoves - phase >= (wmoves + 1) * 7:
        yield wmoves - bmoves, _("White has a rather cramped position")
    elif bmoves - phase >= (wmoves + 1) * 3:
        yield wmoves - bmoves, _("White has a slightly cramped position")



import asyncio
import shutil
import os
import stat

from gi.repository import Gtk

# from pychess.ic import TimeSeal
from pychess.Savers import pgn
from pychess.System import conf
from pychess.System import uistuff
from pychess.System import download_file_async
from pychess.System.prefix import getEngineDataPrefix
from pychess.widgets import mainwindow


class ExternalsDialog:
    def __init__(self):
        self.window = Gtk.Window(
            Gtk.WindowType.TOPLEVEL, title=_("Ask for permissions")
        )
        self.window.set_transient_for(mainwindow())
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        gtk_version = (Gtk.get_major_version(), Gtk.get_minor_version())
        if gtk_version >= (3, 12):
            vbox.props.margin_start = 9
            vbox.props.margin_end = 9
        else:
            vbox.props.margin_left = 9
            vbox.props.margin_right = 9
        vbox.props.margin_bottom = 9
        self.window.add(vbox)
        uistuff.keepWindowSize(
            "externalsdialog", self.window, (320, 240), uistuff.POSITION_CENTER
        )

        label = Gtk.Label(
            _(
                "Some PyChess features require further permission in order to download external components"
            )
        )
        vbox.pack_start(label, True, True, 0)

        box = Gtk.Box()
        check_button = Gtk.CheckButton(_("database querying needs scoutfish"))
        check_button.set_active(conf.get("download_scoutfish"))
        check_button.connect(
            "toggled", lambda w: conf.set("download_scoutfish", w.get_active())
        )
        box.add(check_button)
        link = "https://github.com/pychess/scoutfish"
        link_button = Gtk.LinkButton(link, link)
        box.add(link_button)
        vbox.pack_start(box, False, False, 0)

        box = Gtk.Box()
        check_button = Gtk.CheckButton(_("database opening tree needs chess_db"))
        check_button.set_active(conf.get("download_chess_db"))
        check_button.connect(
            "toggled", lambda w: conf.set("download_chess_db", w.get_active())
        )
        box.add(check_button)
        link = "https://github.com/pychess/chess_db"
        link_button = Gtk.LinkButton(link, link)
        box.add(link_button)
        vbox.pack_start(box, False, False, 0)

        # ICC doesn't support telnet connection any more
        # See https://github.com/pychess/pychess/issues/2315

        # box = Gtk.Box()
        # check_button = Gtk.CheckButton(_("ICC lag compensation needs timestamp"))
        # check_button.set_active(conf.get("download_timestamp"))
        # check_button.connect(
        # "toggled", lambda w: conf.set("download_timestamp", w.get_active())
        # )
        # box.add(check_button)
        # link = "http://download.chessclub.com/timestamp/"
        # link_button = Gtk.LinkButton(link, link)
        # box.add(link_button)
        # vbox.pack_start(box, False, False, 0)

        check_button = Gtk.CheckButton(_("Don't show this dialog on startup."))
        check_button.set_active(conf.get("dont_show_externals_at_startup"))
        check_button.connect(
            "toggled",
            lambda w: conf.set("dont_show_externals_at_startup", w.get_active()),
        )
        vbox.pack_start(check_button, True, True, 0)

        buttonbox = Gtk.ButtonBox()
        close_button = Gtk.Button.new_from_stock(Gtk.STOCK_OK)
        close_button.connect("clicked", self.on_close_clicked)
        self.window.connect("delete_event", lambda w, a: self.window.destroy())
        buttonbox.add(close_button)
        vbox.pack_start(buttonbox, False, False, 0)

    def show(self):
        self.window.show_all()
        self.window.present()

    def on_close_clicked(self, button):
        async def coro():
            altpath = getEngineDataPrefix()
            if pgn.scoutfish_path is None and conf.get("download_scoutfish"):
                binary = (
                    "https://github.com/pychess/scoutfish/releases/download/20170627/%s"
                    % pgn.scoutfish
                )
                filename = await download_file_async(binary)
                if filename is not None:
                    dest = shutil.move(filename, os.path.join(altpath, pgn.scoutfish))
                    os.chmod(dest, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)
                    pgn.scoutfish_path = dest

            if pgn.chess_db_path is None and conf.get("download_chess_db"):
                binary = (
                    "https://github.com/pychess/chess_db/releases/download/20170627/%s"
                    % pgn.parser
                )
                filename = await download_file_async(binary)
                if filename is not None:
                    dest = shutil.move(filename, os.path.join(altpath, pgn.parser))
                    os.chmod(dest, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)
                    pgn.chess_db_path = dest

            # ICC doesn't support telnet connection any more
            # See https://github.com/pychess/pychess/issues/2315

            # if TimeSeal.timestamp_path is None and conf.get("download_timestamp"):
            # binary = (
            # "http://download.chessclub.com.s3.amazonaws.com/timestamp/%s"
            # % TimeSeal.timestamp
            # )
            # filename = await download_file_async(binary)
            # if filename is not None:
            # dest = shutil.move(
            # filename, os.path.join(altpath, TimeSeal.timestamp)
            # )
            # os.chmod(dest, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)
            # TimeSeal.timestamp_path = dest

        asyncio.create_task(coro())

        self.window.emit("delete-event", None)



import gc
import types

# from pychess.Utils.Board import Board
from pychess.Utils.GameModel import GameModel

# from pychess.Utils.lutils.LBoard import LBoard
from pychess.widgets.BoardView import BoardView
from pychess.widgets.BoardControl import BoardControl
from pychess.widgets.gamewidget import GameWidget
from pychess.widgets.pydock.PyDockTop import PyDockTop
from pychess.widgets.pydock.PyDockLeaf import PyDockLeaf

# from pychess.widgets.pydock.ArrowButton import ArrowButton
# from pychess.widgets.pydock.StarArrowButton import StarArrowButton
# from pychess.widgets.pydock.HighlightArea import HighlightArea
from pychess.Players.CECPEngine import CECPEngine
from pychess.Players.UCIEngine import UCIEngine
from pychess.Players.Human import Human
from pychess.Players.ICPlayer import ICPlayer
from pychess.ic.ICGameModel import ICGameModel
from pychess.ic import ICLogon
from pychess.perspectives import perspective_manager


def obj_referrers(klass):
    find_obj = False
    for obj in gc.get_objects():
        # closures are evil !
        if isinstance(obj, types.FunctionType) and obj.__closure__ is not None:
            for c in obj.__closure__:
                try:
                    if isinstance(c.cell_contents, klass):
                        print("!!!", obj, c.cell_contents)
                except ValueError:
                    print("Cell is empty...")
        if isinstance(obj, klass):
            find_obj = True
            rs = gc.get_referrers(obj)
            print("---------------------------referrers of %s" % klass.__name__)
            for ob in rs:
                print(type(ob), ob.__name__ if type(ob) is type else repr(ob)[:140])
                rs1 = gc.get_referrers(ob)
                for ob1 in rs1:
                    print(
                        "    ",
                        type(ob1),
                        ob1.__name__ if type(ob1) is type else repr(ob1)[:140],
                    )
            print("---------------------------")
    if not find_obj:
        print("Nothing refrences %s" % klass.__name__)


def print_obj_referrers():
    perspective = perspective_manager.get_perspective("games")
    if len(perspective.gamewidgets) > 0:
        return

    for klass in (
        ICGameModel,
        GameModel,
        GameWidget,
        BoardView,
        BoardControl,
        CECPEngine,
        UCIEngine,
        Human,
        ICPlayer,
        # TODO:
        # ArrowButton,
        # StarArrowButton,
        # HighlightArea,
        # Board,
        # LBoard,
    ):
        obj_referrers(klass)

    if ICLogon.dialog is None or not hasattr(ICLogon.dialog, "lounge"):
        for klass in (
            PyDockTop,
            PyDockLeaf,
        ):
            obj_referrers(klass)

    print("---------------------------------")


def print_muppy_sumary():
    # http://pythonhosted.org/Pympler/index.html
    try:
        from pympler import muppy, summary
    except ImportError:
        print("WARNING: pympler not installed")
        return
    # from pympler.classtracker import ClassTracker
    # from pympler.classtracker_stats import HtmlStats
    global all_objects, obj_summary, class_tracker
    if all_objects is None:
        all_objects = muppy.get_objects()
        obj_summary = summary.summarize(all_objects)
        summary.print_(obj_summary)

        # class_tracker = ClassTracker()
        # class_tracker.track_class(FICSPlayer, trace=1)
        # class_tracker.track_class(ICGameModel, resolution_level=2, trace=1)
    else:
        obj_summary2 = summary.summarize(muppy.get_objects())
        diff = summary.get_diff(obj_summary, obj_summary2)
        summary.print_(diff, limit=200)

        # class_tracker.create_snapshot('usage')
        # HtmlStats(tracker=class_tracker).create_html('profile.html')


all_objects = None
obj_summary = None
class_tracker = None
